@@ -66,6 +66,7 @@ meta_core_get_frame_type (Display *xdisplay,
 {
   MetaDisplay *display;
   MetaWindow *window;
+  MetaFrameType base_type;
   
   display = meta_display_for_x_display (xdisplay);
   window = meta_display_lookup_x_window (display, frame_xwindow);
@@ -73,26 +74,28 @@ meta_core_get_frame_type (Display *xdisplay,
   if (window == NULL || window->frame == NULL)
     meta_bug ("No such frame window 0x%lx!\n", frame_xwindow);
 
+  base_type = META_FRAME_TYPE_LAST;
+  
   switch (window->type)
     {
     case META_WINDOW_NORMAL:
-      return META_FRAME_TYPE_NORMAL;
+      base_type = META_FRAME_TYPE_NORMAL;
       break;
       
     case META_WINDOW_DIALOG:
-      return META_FRAME_TYPE_DIALOG;
+      base_type = META_FRAME_TYPE_DIALOG;
       break;
       
     case META_WINDOW_MODAL_DIALOG:
-      return META_FRAME_TYPE_MODAL_DIALOG;
+      base_type = META_FRAME_TYPE_MODAL_DIALOG;
       break;
       
     case META_WINDOW_MENU:
-      return META_FRAME_TYPE_MENU;
+      base_type = META_FRAME_TYPE_MENU;
       break;
 
     case META_WINDOW_UTILITY:
-      return META_FRAME_TYPE_UTILITY;
+      base_type = META_FRAME_TYPE_UTILITY;
       break;
       
     case META_WINDOW_DESKTOP:
@@ -100,12 +103,16 @@ meta_core_get_frame_type (Display *xdisplay,
     case META_WINDOW_TOOLBAR:
     case META_WINDOW_SPLASHSCREEN:
       /* No frame */
-      return META_FRAME_TYPE_LAST;
+      base_type = META_FRAME_TYPE_LAST;
       break;
     }
 
-  g_assert_not_reached ();
-  return META_FRAME_TYPE_LAST;
+  if (base_type == META_FRAME_TYPE_LAST)
+    return META_FRAME_TYPE_LAST; /* can't add border if undecorated */
+  else if (window->border_only)
+    return META_FRAME_TYPE_BORDER; /* override base frame type */
+  else
+    return base_type;
 }
 
 GdkPixbuf*
@@ -451,6 +458,34 @@ meta_core_get_frame_workspace (Display *xdisplay,
   return meta_window_get_net_wm_desktop (window);
 }
 
+void
+meta_core_get_frame_extents   (Display        *xdisplay,
+                               Window          frame_xwindow,
+                               int            *x,
+                               int            *y,
+                               int            *width,
+                               int            *height)
+{
+  MetaDisplay *display;
+  MetaWindow *window;
+  
+  display = meta_display_for_x_display (xdisplay);
+  window = meta_display_lookup_x_window (display, frame_xwindow);
+
+  if (window == NULL || window->frame == NULL)
+    meta_bug ("No such frame window 0x%lx!\n", frame_xwindow);  
+
+  if (x)
+    *x = window->frame->rect.x;
+  if (y)
+    *y = window->frame->rect.y;
+  if (width)
+    *width = window->frame->rect.width;
+  if (height)
+    *height = window->frame->rect.height;
+}
+
+
 void
 meta_core_show_window_menu (Display *xdisplay,
                             Window   frame_xwindow,
@@ -471,6 +506,18 @@ meta_core_show_window_menu (Display *xdisplay,
   meta_window_show_menu (window, root_x, root_y, button, timestamp);
 }
 
+char *
+meta_core_get_workspace_name_with_index (Display *xdisplay,
+                                         int index)
+{
+  MetaDisplay *display;
+  MetaWorkspace *workspace;
+
+  display = meta_display_for_x_display (xdisplay);
+  workspace = meta_display_get_workspace_by_index (display, index);
+  return (workspace != NULL) ? workspace->name : NULL;
+}
+
 gboolean
 meta_core_begin_grab_op (Display    *xdisplay,
                          Window      frame_xwindow,
@@ -484,14 +531,18 @@ meta_core_begin_grab_op (Display    *xdisplay,
 {
   MetaDisplay *display;
   MetaWindow *window;
+  MetaScreen *screen;
   
   display = meta_display_for_x_display (xdisplay);
+  screen = meta_display_screen_for_xwindow (display, frame_xwindow);
   window = meta_display_lookup_x_window (display, frame_xwindow);
 
+  g_assert (screen != NULL);
+  
   if (window == NULL || window->frame == NULL)
     meta_bug ("No such frame window 0x%lx!\n", frame_xwindow);  
 
-  return meta_display_begin_grab_op (display, window,
+  return meta_display_begin_grab_op (display, screen, window,
                                      op, pointer_already_grabbed,
                                      button, modmask,
                                      timestamp, root_x, root_y);
@@ -527,11 +578,12 @@ meta_core_get_grab_frame (Display *xdisplay)
 
   g_assert (display != NULL);
   g_assert (display->grab_op == META_GRAB_OP_NONE || 
-            display->grab_window != NULL);
+            display->grab_screen != NULL);
   g_assert (display->grab_op == META_GRAB_OP_NONE ||
-            display->grab_window->display->xdisplay == xdisplay);
+            display->grab_screen->display->xdisplay == xdisplay);
   
   if (display->grab_op != META_GRAB_OP_NONE &&
+      display->grab_window &&
       display->grab_window->frame)
     return display->grab_window->frame->xwindow;
   else
@@ -558,7 +610,8 @@ meta_core_grab_buttons  (Display *xdisplay,
   MetaDisplay *display;
   
   display = meta_display_for_x_display (xdisplay);
-  
+
+  meta_verbose ("Grabbing buttons on frame 0x%lx\n", frame_xwindow);
   meta_display_grab_window_buttons (display, frame_xwindow);
 }
 
@@ -579,6 +632,27 @@ meta_core_set_screen_cursor (Display *xdisplay,
   meta_screen_set_cursor (window->screen, cursor);
 }
 
+void
+meta_core_get_screen_size (Display *xdisplay,
+                           Window   frame_on_screen,
+                           int     *width,
+                           int     *height)
+{
+  MetaDisplay *display;
+  MetaWindow *window;
+  
+  display = meta_display_for_x_display (xdisplay);
+  window = meta_display_lookup_x_window (display, frame_on_screen);
+
+  if (window == NULL || window->frame == NULL)
+    meta_bug ("No such frame window 0x%lx!\n", frame_on_screen);  
+
+  if (width)
+    *width = window->screen->width;
+  if (height)
+    *height = window->screen->height;
+}
+
 void
 meta_core_increment_event_serial (Display *xdisplay)
 {
