@@ -98,6 +98,16 @@ int meta_core_get_num_workspaces (Screen  *xscreen);
 int meta_core_get_active_workspace (Screen *xscreen);
 int meta_core_get_frame_workspace (Display *xdisplay,
                                    Window frame_xwindow);
+char* meta_core_get_workspace_name_with_index (Display *xdisplay,
+                                               int index);
+
+void  meta_core_get_frame_extents   (Display        *xdisplay,
+                                     Window          frame_xwindow,
+                                     int            *x,
+                                     int            *y,
+                                     int            *width,
+                                     int            *height);
+
 
 void meta_core_show_window_menu (Display *xdisplay,
                                  Window   frame_xwindow,
@@ -129,6 +139,11 @@ void       meta_core_set_screen_cursor (Display *xdisplay,
                                         Window   frame_on_screen,
                                         MetaCursor cursor);
 
+void       meta_core_get_screen_size (Display *xdisplay,
+                                      Window   frame_on_screen,
+                                      int     *width,
+                                      int     *height);
+
 /* Used because we ignore EnterNotify when a window is unmapped that
  * really shouldn't cause focus changes, by comparing the event serial
  * of the EnterNotify and the UnmapNotify.
