@@ -40,13 +40,14 @@ struct _MetaRectangle
   int height;
 };
 
-typedef struct _MetaDisplay   MetaDisplay;
-typedef struct _MetaFrame     MetaFrame;
-typedef struct _MetaScreen    MetaScreen;
-typedef struct _MetaStack     MetaStack;
-typedef struct _MetaUISlave   MetaUISlave;
-typedef struct _MetaWindow    MetaWindow;
-typedef struct _MetaWorkspace MetaWorkspace;
+typedef struct _MetaDisplay    MetaDisplay;
+typedef struct _MetaFrame      MetaFrame;
+typedef struct _MetaKeyBinding MetaKeyBinding;
+typedef struct _MetaScreen     MetaScreen;
+typedef struct _MetaStack      MetaStack;
+typedef struct _MetaUISlave    MetaUISlave;
+typedef struct _MetaWindow     MetaWindow;
+typedef struct _MetaWorkspace  MetaWorkspace;
 
 typedef void (* MetaWindowPingFunc) (MetaDisplay *display,
 				     Window       xwindow,
@@ -127,6 +128,25 @@ struct _MetaDisplay
   Atom atom_net_wm_pid;
   Atom atom_wm_client_machine;
   Atom atom_net_wm_state_fullscreen;
+  Atom atom_net_workarea;
+  Atom atom_net_show_desktop;
+  Atom atom_net_desktop_layout;
+  Atom atom_manager;
+  Atom atom_targets;
+  Atom atom_multiple;
+  Atom atom_timestamp;
+  Atom atom_version;
+  Atom atom_atom_pair;
+  Atom atom_net_desktop_names;
+  Atom atom_net_wm_allowed_actions;
+  Atom atom_net_wm_action_move;
+  Atom atom_net_wm_action_resize;
+  Atom atom_net_wm_action_shade;
+  Atom atom_net_wm_action_stick;
+  Atom atom_net_wm_action_maximize_horz;
+  Atom atom_net_wm_action_maximize_vert;
+  Atom atom_net_wm_action_change_desktop;
+  Atom atom_net_wm_action_close;
   
   /* This is the actual window from focus events,
    * not the one we last set
@@ -175,9 +195,13 @@ struct _MetaDisplay
 
   /* Pings which we're waiting for a reply from */
   GSList     *pending_pings;
+
+  /* Pending autoraise */
+  guint       autoraise_timeout_id;
   
   /* current window operation */
   MetaGrabOp  grab_op;
+  MetaScreen *grab_screen;
   MetaWindow *grab_window;
   Window      grab_xwindow;
   int         grab_button;
@@ -188,6 +212,30 @@ struct _MetaDisplay
   guint       grab_have_keyboard : 1;
   MetaRectangle grab_initial_window_pos;
   MetaResizePopup *grab_resize_popup;
+
+
+  /* Keybindings stuff */
+  MetaKeyBinding *screen_bindings;
+  int             n_screen_bindings;
+  MetaKeyBinding *window_bindings;
+  int             n_window_bindings;
+  int min_keycode;
+  int max_keycode;
+  KeySym *keymap;
+  int keysyms_per_keycode;
+  XModifierKeymap *modmap;
+  unsigned int ignored_modifier_mask;
+  unsigned int num_lock_mask;
+  unsigned int scroll_lock_mask;
+  unsigned int hyper_mask;
+  unsigned int super_mask;
+  unsigned int meta_mask;
+  
+  /* Xinerama cache */
+  unsigned int xinerama_cache_invalidated : 1;
+
+  /* Closing down the display */
+  int closing;
 };
 
 gboolean      meta_display_open                (const char  *name);
@@ -196,10 +244,18 @@ MetaScreen*   meta_display_screen_for_root     (MetaDisplay *display,
                                                 Window       xroot);
 MetaScreen*   meta_display_screen_for_x_screen (MetaDisplay *display,
                                                 Screen      *screen);
+MetaScreen*   meta_display_screen_for_xwindow  (MetaDisplay *display,
+                                                Window       xindow);
 void          meta_display_grab                (MetaDisplay *display);
 void          meta_display_ungrab              (MetaDisplay *display);
 gboolean      meta_display_is_double_click     (MetaDisplay *display);
 
+void          meta_display_unmanage_screen     (MetaDisplay *display,
+                                                MetaScreen  *screen);
+
+void          meta_display_unmanage_windows_for_screen (MetaDisplay *display,
+                                                        MetaScreen  *screen);
+
 /* A given MetaWindow may have various X windows that "belong"
  * to it, such as the frame window.
  */
@@ -226,6 +282,7 @@ Cursor         meta_display_create_x_cursor (MetaDisplay *display,
                                              MetaCursor   cursor);
 
 gboolean meta_display_begin_grab_op (MetaDisplay *display,
+                                     MetaScreen  *screen,
                                      MetaWindow  *window,
                                      MetaGrabOp   op,
                                      gboolean     pointer_already_grabbed,
@@ -289,10 +346,15 @@ GSList* meta_display_get_tab_list (MetaDisplay   *display,
 
 MetaWindow* meta_display_get_tab_next (MetaDisplay   *display,
                                        MetaTabList    type,
+				       MetaScreen    *screen,
                                        MetaWorkspace *workspace,
                                        MetaWindow    *window,
                                        gboolean       backward);
 
 int meta_resize_gravity_from_grab_op (MetaGrabOp op);
 
+gboolean meta_rectangle_intersect (MetaRectangle *src1,
+                                   MetaRectangle *src2,
+                                   MetaRectangle *dest);
+
 #endif
