@@ -26,6 +26,11 @@
 #include "menu.h"
 #include "fixedtip.h"
 #include "theme.h"
+#include "prefs.h"
+
+#ifdef HAVE_SHAPE
+#include <X11/extensions/shape.h>
+#endif
 
 #define DEFAULT_INNER_BUTTON_BORDER 3
 
@@ -56,7 +61,7 @@ static gboolean meta_frames_leave_notify_event    (GtkWidget           *widget,
 static void meta_frames_paint_to_drawable (MetaFrames   *frames,
                                            MetaUIFrame  *frame,
                                            GdkDrawable  *drawable,
-                                           GdkRectangle *area);
+                                           GdkRegion    *region);
 
 static void meta_frames_calc_geometry (MetaFrames        *frames,
                                        MetaUIFrame         *frame,
@@ -68,6 +73,7 @@ static void meta_frames_ensure_layout (MetaFrames      *frames,
 static MetaUIFrame* meta_frames_lookup_window (MetaFrames *frames,
                                                Window      xwindow);
 
+static void meta_frames_font_changed (MetaFrames *frames);
 
 static GdkRectangle*    control_rect (MetaFrameControl   control,
                                       MetaFrameGeometry *fgeom);
@@ -154,6 +160,16 @@ unsigned_long_hash (gconstpointer v)
 #endif
 }
 
+static void
+font_changed_callback (MetaPreference pref,
+		       void          *data)
+{
+  if (pref == META_PREF_TITLEBAR_FONT)
+    {
+      meta_frames_font_changed (META_FRAMES (data));
+    }
+}
+
 static void
 meta_frames_init (MetaFrames *frames)
 {
@@ -166,6 +182,10 @@ meta_frames_init (MetaFrames *frames)
   frames->tooltip_timeout = 0;
 
   frames->expose_delay_count = 0;
+
+  gtk_widget_set_double_buffered (GTK_WIDGET (frames), FALSE);
+
+  meta_prefs_add_listener (font_changed_callback, frames);
 }
 
 static void
@@ -217,6 +237,8 @@ meta_frames_finalize (GObject *object)
   
   frames = META_FRAMES (object);
 
+  meta_prefs_remove_listener (font_changed_callback, frames);
+  
   g_hash_table_destroy (frames->text_heights);
   
   g_assert (g_hash_table_size (frames->frames) == 0);
@@ -256,13 +278,8 @@ queue_recalc_func (gpointer key, gpointer value, gpointer data)
 }
 
 static void
-meta_frames_style_set  (GtkWidget *widget,
-                        GtkStyle  *prev_style)
+meta_frames_font_changed (MetaFrames *frames)
 {
-  MetaFrames *frames;
-
-  frames = META_FRAMES (widget);
-
   if (g_hash_table_size (frames->text_heights) > 0)
     {
       g_hash_table_destroy (frames->text_heights);
@@ -273,6 +290,18 @@ meta_frames_style_set  (GtkWidget *widget,
   g_hash_table_foreach (frames->frames,
                         queue_recalc_func, frames);
 
+}
+
+static void
+meta_frames_style_set  (GtkWidget *widget,
+                        GtkStyle  *prev_style)
+{
+  MetaFrames *frames;
+
+  frames = META_FRAMES (widget);
+
+  meta_frames_font_changed (frames);
+
   GTK_WIDGET_CLASS (parent_class)->style_set (widget, prev_style);
 }
 
@@ -324,7 +353,8 @@ meta_frames_ensure_layout (MetaFrames  *frames,
       
       frame->layout = gtk_widget_create_pango_layout (widget, frame->title);
       
-      font_desc = meta_gtk_widget_get_font_desc (widget, scale);
+      font_desc = meta_gtk_widget_get_font_desc (widget, scale,
+						 meta_prefs_get_titlebar_font ());
 
       size = pango_font_description_get_size (font_desc);
       
@@ -344,25 +374,9 @@ meta_frames_ensure_layout (MetaFrames  *frames,
                                 &size,
                                 GINT_TO_POINTER (frame->text_height));
         }
-
-      if (pango_font_description_get_size (font_desc) !=
-          pango_font_description_get_size (widget->style->font_desc))
-        {
-          PangoAttrList *attrs;
-          PangoAttribute *attr;
-          
-          attrs = pango_attr_list_new ();
-          
-          attr = pango_attr_size_new (pango_font_description_get_size (font_desc));
-          attr->start_index = 0;
-          attr->end_index = G_MAXINT;
-
-          pango_attr_list_insert (attrs, attr);
-
-          pango_layout_set_attributes (frame->layout, attrs);
-
-          pango_attr_list_unref (attrs);
-        }
+      
+      pango_layout_set_font_description (frame->layout, 
+					 font_desc);
       
       pango_font_description_free (font_desc);
 
@@ -398,9 +412,22 @@ meta_frames_calc_geometry (MetaFrames        *frames,
 }
 
 MetaFrames*
-meta_frames_new (void)
+meta_frames_new (int screen_number)
 {
-  return g_object_new (META_TYPE_FRAMES, NULL);
+#ifdef HAVE_GTK_MULTIHEAD
+  GdkScreen *screen;
+
+  screen = gdk_display_get_screen (gdk_display_get_default (),
+				   screen_number);
+
+  return g_object_new (META_TYPE_FRAMES,
+		       "screen", screen,
+		       NULL);
+#else
+  return g_object_new (META_TYPE_FRAMES,
+		       NULL);
+#endif
+  
 }
 
 void
@@ -448,7 +475,12 @@ meta_frames_unmanage_window (MetaFrames *frames,
   frame = g_hash_table_lookup (frames->frames, &xwindow);
 
   if (frame)
-    {      
+    {
+      /* restore the cursor */
+      meta_core_set_screen_cursor (gdk_display,
+				   frame->xwindow,
+				   META_CURSOR_DEFAULT);
+
       gdk_window_set_user_data (frame->window, NULL);
 
       if (frames->last_motion_frame == frame)
@@ -590,6 +622,158 @@ meta_frames_unflicker_bg (MetaFrames *frames,
   set_background_none (gdk_display, frame->xwindow);
 }
 
+void
+meta_frames_apply_shapes (MetaFrames *frames,
+                          Window      xwindow,
+                          int         new_window_width,
+                          int         new_window_height)
+{
+#ifdef HAVE_SHAPE
+  /* Apply shapes as if window had new_window_width, new_window_height */
+  GtkWidget *widget;
+  MetaUIFrame *frame;
+  MetaFrameGeometry fgeom;
+  XRectangle xrect;
+  Region corners_xregion;
+  Region window_xregion;
+  
+  widget = GTK_WIDGET (frames);
+
+  frame = meta_frames_lookup_window (frames, xwindow);
+  g_return_if_fail (frame != NULL);
+
+  meta_frames_calc_geometry (frames, frame, &fgeom);
+
+  if (!(fgeom.top_left_corner_rounded ||
+        fgeom.top_right_corner_rounded ||
+        fgeom.bottom_left_corner_rounded ||
+        fgeom.bottom_right_corner_rounded))
+    {
+      XShapeCombineMask (gdk_display, frame->xwindow,
+                         ShapeBounding, 0, 0, None, ShapeSet);
+      
+      return; /* nothing to do */
+    }
+  
+  corners_xregion = XCreateRegion ();
+  
+  if (fgeom.top_left_corner_rounded)
+    {
+      xrect.x = 0;
+      xrect.y = 0;
+      xrect.width = 5;
+      xrect.height = 1;
+      
+      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+
+      xrect.y = 1;
+      xrect.width = 3;
+      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+      
+      xrect.y = 2;
+      xrect.width = 2;
+      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+
+      xrect.y = 3;
+      xrect.width = 1;
+      xrect.height = 2;
+      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+    }
+
+  if (fgeom.top_right_corner_rounded)
+    {
+      xrect.x = new_window_width - 5;
+      xrect.y = 0;
+      xrect.width = 5;
+      xrect.height = 1;
+      
+      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+
+      xrect.y = 1;
+      xrect.x = new_window_width - 3;
+      xrect.width = 3;
+      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+      
+      xrect.y = 2;
+      xrect.x = new_window_width - 2;
+      xrect.width = 2;
+      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+
+      xrect.y = 3;
+      xrect.x = new_window_width - 1;
+      xrect.width = 1;
+      xrect.height = 2;
+      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+    }
+
+  if (fgeom.bottom_left_corner_rounded)
+    {
+      xrect.x = 0;
+      xrect.y = new_window_height - 1;
+      xrect.width = 5;
+      xrect.height = 1;
+      
+      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+      
+      xrect.y = new_window_height - 2;
+      xrect.width = 3;
+      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+      
+      xrect.y = new_window_height - 3;
+      xrect.width = 2;
+      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+
+      xrect.y = new_window_height - 5;
+      xrect.width = 1;
+      xrect.height = 2;
+      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+    }
+
+  if (fgeom.bottom_right_corner_rounded)
+    {
+      xrect.x = new_window_width - 5;
+      xrect.y = new_window_height - 1;
+      xrect.width = 5;
+      xrect.height = 1;
+      
+      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+      
+      xrect.y = new_window_height - 2;
+      xrect.x = new_window_width - 3;
+      xrect.width = 3;
+      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+      
+      xrect.y = new_window_height - 3;
+      xrect.x = new_window_width - 2;
+      xrect.width = 2;
+      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+
+      xrect.y = new_window_height - 5;
+      xrect.x = new_window_width - 1;
+      xrect.width = 1;
+      xrect.height = 2;
+      XUnionRectWithRegion (&xrect, corners_xregion, corners_xregion);
+    }
+  
+  window_xregion = XCreateRegion ();
+  
+  xrect.x = 0;
+  xrect.y = 0;
+  xrect.width = new_window_width;
+  xrect.height = new_window_height;
+
+  XUnionRectWithRegion (&xrect, window_xregion, window_xregion);
+
+  XSubtractRegion (window_xregion, corners_xregion, window_xregion);
+
+  XShapeCombineRegion (gdk_display, frame->xwindow,
+                       ShapeBounding, 0, 0, window_xregion, ShapeSet);
+  
+  XDestroyRegion (window_xregion);
+  XDestroyRegion (corners_xregion);
+#endif
+}
+
 void
 meta_frames_queue_draw (MetaFrames *frames,
                         Window      xwindow)
@@ -616,6 +800,8 @@ meta_frames_set_title (MetaFrames *frames,
 
   frame = meta_frames_lookup_window (frames, xwindow);
 
+  g_assert (frame);
+  
   g_free (frame->title);
   frame->title = g_strdup (title);
   
@@ -628,6 +814,21 @@ meta_frames_set_title (MetaFrames *frames,
   gdk_window_invalidate_rect (frame->window, NULL, FALSE);
 }
 
+void
+meta_frames_repaint_frame (MetaFrames *frames,
+                           Window      xwindow)
+{
+  GtkWidget *widget;
+  MetaUIFrame *frame;
+  
+  widget = GTK_WIDGET (frames);
+
+  frame = meta_frames_lookup_window (frames, xwindow);
+
+  g_assert (frame);
+
+  gdk_window_process_updates (frame->window, TRUE);
+}
 
 static void
 show_tip_now (MetaFrames *frames)
@@ -699,6 +900,7 @@ show_tip_now (MetaFrames *frames)
       MetaFrameGeometry fgeom;
       GdkRectangle *rect;
       int dx, dy;
+      int screen_number;
       
       meta_frames_calc_geometry (frames, frame, &fgeom);
       
@@ -707,8 +909,13 @@ show_tip_now (MetaFrames *frames)
       /* get conversion delta for root-to-frame coords */
       dx = root_x - x;
       dy = root_y - y;
-      
+#ifdef HAVE_GTK_MULTIHEAD
+      screen_number = gdk_screen_get_number (gtk_widget_get_screen (GTK_WIDGET (frames)));
+#else
+      screen_number = DefaultScreen (gdk_display);
+#endif
       meta_fixed_tip_show (gdk_display,
+			   screen_number,
                            rect->x + dx,
                            rect->y + rect->height + 2 + dy,
                            tiptext);
@@ -795,10 +1002,10 @@ meta_frames_button_press_event (GtkWidget      *widget,
   
   control = get_control (frames, frame, event->x, event->y);
 
-  if (control == META_FRAME_CONTROL_CLIENT_AREA)
-    return FALSE; /* not on the frame, just passed through from client */
-
-  if (event->button == 1)
+  /* raise/focus on click, even if click was on client area */
+  if (event->button == 1 &&
+      !(control == META_FRAME_CONTROL_MINIMIZE ||
+        control == META_FRAME_CONTROL_DELETE))
     {
       meta_core_user_raise (gdk_display,
                             frame->xwindow);
@@ -809,6 +1016,10 @@ meta_frames_button_press_event (GtkWidget      *widget,
                             frame->xwindow,
                             event->time);      
     }
+
+  /* don't do the rest of this if on client area */
+  if (control == META_FRAME_CONTROL_CLIENT_AREA)
+    return FALSE; /* not on the frame, just passed through from client */
   
   /* We want to shade even if we have a GrabOp, since we'll have a move grab
    * if we double click the titlebar.
@@ -1275,7 +1486,7 @@ meta_frames_expose_event (GtkWidget           *widget,
       return TRUE;
     }
 
-  meta_frames_paint_to_drawable (frames, frame, frame->window, &event->area);
+  meta_frames_paint_to_drawable (frames, frame, frame->window, event->region);
 
   return TRUE;
 }
@@ -1284,7 +1495,7 @@ static void
 meta_frames_paint_to_drawable (MetaFrames   *frames,
                                MetaUIFrame  *frame,
                                GdkDrawable  *drawable,
-                               GdkRectangle *area)
+                               GdkRegion    *region)
 {
   GtkWidget *widget;
   MetaFrameFlags flags;
@@ -1295,6 +1506,13 @@ meta_frames_paint_to_drawable (MetaFrames   *frames,
   MetaButtonState button_states[META_BUTTON_TYPE_LAST];
   Window grab_frame;
   int i;
+  int top, bottom, left, right;
+  GdkRegion *edges;
+  GdkRegion *tmp_region;
+  GdkRectangle area;
+  GdkRectangle *areas;
+  int n_areas;
+  int screen_width, screen_height;
   
   widget = GTK_WIDGET (frames);
 
@@ -1322,6 +1540,7 @@ meta_frames_paint_to_drawable (MetaFrames   *frames,
       button_states[META_BUTTON_TYPE_CLOSE] = META_BUTTON_STATE_PRELIGHT;
       break;
     default:
+      break;
     }
   
   grab_frame = meta_core_get_grab_frame (gdk_display);
@@ -1354,6 +1573,20 @@ meta_frames_paint_to_drawable (MetaFrames   *frames,
           break;
         }
     }
+
+  /* Map button function states to button position states */
+  button_states[META_BUTTON_TYPE_LEFT_LEFT_BACKGROUND] =
+    button_states[META_BUTTON_TYPE_MENU];
+  button_states[META_BUTTON_TYPE_LEFT_MIDDLE_BACKGROUND] =
+    META_BUTTON_STATE_NORMAL;
+  button_states[META_BUTTON_TYPE_LEFT_RIGHT_BACKGROUND] =
+    META_BUTTON_STATE_NORMAL;
+  button_states[META_BUTTON_TYPE_RIGHT_LEFT_BACKGROUND] =
+    button_states[META_BUTTON_TYPE_MINIMIZE];
+  button_states[META_BUTTON_TYPE_RIGHT_MIDDLE_BACKGROUND] =
+    button_states[META_BUTTON_TYPE_MAXIMIZE];
+  button_states[META_BUTTON_TYPE_RIGHT_RIGHT_BACKGROUND] =
+    button_states[META_BUTTON_TYPE_CLOSE];
   
   flags = meta_core_get_frame_flags (gdk_display, frame->xwindow);
   type = meta_core_get_frame_type (gdk_display, frame->xwindow);
@@ -1364,19 +1597,84 @@ meta_frames_paint_to_drawable (MetaFrames   *frames,
                              &w, &h);
 
   meta_frames_ensure_layout (frames, frame);
+
+  meta_theme_get_frame_borders (meta_theme_get_current (),
+                                type, frame->text_height, flags, 
+                                &top, &bottom, &left, &right);
+
+  /* Repaint each side of the frame */
   
-  meta_theme_draw_frame (meta_theme_get_current (),
-                         widget,
-                         drawable,
-                         area,
-                         0, 0,
-                         type,
-                         flags,
-                         w, h,
-                         frame->layout,
-                         frame->text_height,
-                         button_states,
-                         mini_icon, icon);
+  edges = gdk_region_copy (region);
+
+  /* Punch out the client area */
+  area.x = left;
+  area.y = top;
+  area.width = w;
+  area.height = h;
+  tmp_region = gdk_region_rectangle (&area);
+  gdk_region_subtract (edges, tmp_region);
+  gdk_region_destroy (tmp_region);
+
+  /* Chop off stuff outside the screen; this optimization
+   * is crucial to handle huge client windows,
+   * like "xterm -geometry 1000x1000"
+   */
+  meta_core_get_frame_extents (gdk_display,
+                               frame->xwindow,
+                               &area.x, &area.y,
+                               &area.width, &area.height);
+
+  meta_core_get_screen_size (gdk_display,
+                             frame->xwindow,
+                             &screen_width, &screen_height);
+
+  if ((area.x + area.width) > screen_width)
+    area.width = screen_width - area.x;
+  if (area.width < 0)
+    area.width = 0;
+  
+  if ((area.y + area.height) > screen_height)
+    area.height = screen_height - area.y;
+  if (area.height < 0)
+    area.height = 0;
+
+  area.x = 0; /* make relative to frame rather than screen */
+  area.y = 0;
+  
+  tmp_region = gdk_region_rectangle (&area);
+  gdk_region_intersect (edges, tmp_region);
+  gdk_region_destroy (tmp_region);
+
+  /* Now draw remaining portion of region */
+  gdk_region_get_rectangles (edges, &areas, &n_areas);
+  
+  i = 0;
+  while (i < n_areas)
+    {      
+      if (GDK_IS_WINDOW (drawable))
+        gdk_window_begin_paint_rect (drawable, &areas[i]);
+      
+      meta_theme_draw_frame (meta_theme_get_current (),
+                             widget,
+                             drawable,
+                             &areas[i],
+                             0, 0,
+                             type,
+                             flags,
+                             w, h,
+                             frame->layout,
+                             frame->text_height,
+                             button_states,
+                             mini_icon, icon);
+
+      if (GDK_IS_WINDOW (drawable))
+        gdk_window_end_paint (drawable);
+      
+      ++i;
+    }
+
+  gdk_region_destroy (edges);
+  g_free (areas);
 }
 
 static gboolean
@@ -1473,13 +1771,8 @@ control_rect (MetaFrameControl control,
   return rect;
 }
 
-#define POINT_IN_RECT(xcoord, ycoord, rect) \
- ((xcoord) >= (rect).x &&                   \
-  (xcoord) <  ((rect).x + (rect).width) &&  \
-  (ycoord) >= (rect).y &&                   \
-  (ycoord) <  ((rect).y + (rect).height))
-
 #define RESIZE_EXTENDS 15
+#define TOP_RESIZE_HEIGHT 2
 static MetaFrameControl
 get_control (MetaFrames *frames,
              MetaUIFrame *frame,
@@ -1530,9 +1823,6 @@ get_control (MetaFrames *frames,
       int bottom_of_titlebar;
 
       bottom_of_titlebar = fgeom.title_rect.y + fgeom.title_rect.height;
-
-      if (y < bottom_of_titlebar)
-        goto noresize;
       
       /* South resize always has priority over north resize,
        * in case of overlap.
@@ -1583,7 +1873,7 @@ get_control (MetaFrames *frames,
           if (has_vert)
             return META_FRAME_CONTROL_RESIZE_S;
         }
-      else if (y >= bottom_of_titlebar && y < fgeom.top_height)
+      else if (y < TOP_RESIZE_HEIGHT)
         {
           if (has_vert)
             return META_FRAME_CONTROL_RESIZE_N;
@@ -1599,8 +1889,6 @@ get_control (MetaFrames *frames,
             return META_FRAME_CONTROL_RESIZE_E;
         }
     }
-
- noresize:
   
   return META_FRAME_CONTROL_NONE;
 }
