@@ -19,8 +19,11 @@
  * 02111-1307, USA.
  */
 
+#include <config.h>
+
 #include "place.h"
 #include "workspace.h"
+#include "prefs.h"
 #include <gdk/gdkregion.h>
 #include <math.h>
 #include <stdlib.h>
@@ -100,7 +103,11 @@ find_next_cascade (MetaWindow *window,
    * cascade_x, cascade_y are the target position
    * of NW corner of window frame.
    */
-  meta_window_get_work_area (window, &work_area);
+  /* FIXME this is bogus because we get the current xinerama
+   * for the window based on its position, but we haven't
+   * placed it yet.
+   */
+  meta_window_get_work_area (window, TRUE, &work_area);
   
   cascade_x = MAX (0, work_area.x);
   cascade_y = MAX (0, work_area.y);
@@ -112,13 +119,13 @@ find_next_cascade (MetaWindow *window,
    */
   if (fgeom)
     {
-      x_threshold = MAX (fgeom->left_width, 10);
-      y_threshold = MAX (fgeom->top_height, 10);
+      x_threshold = MAX (fgeom->left_width, 15);
+      y_threshold = MAX (fgeom->top_height, 15);
     }
   else
     {
-      x_threshold = 10;
-      y_threshold = 10;
+      x_threshold = 15;
+      y_threshold = 15;
     }
   
   tmp = sorted;
@@ -153,16 +160,33 @@ find_next_cascade (MetaWindow *window,
           cascade_y = wy;
         }
       else
-        goto found; /* no window at this cascade point. */
+        {
+          /* Keep searching for a further-down-the-diagonal window. */
+        }
         
       tmp = tmp->next;
     }
 
-  /* cascade_x and cascade_y will match the last window in the list. */
+  /* cascade_x and cascade_y will match the last window in the list
+   * that was "in the way" (in the approximate cascade diagonal)
+   */
   
- found:
   g_list_free (sorted);
 
+  /* don't place windows off the screen */
+  if (((cascade_x + (window->frame ? window->frame->rect.width : window->rect.width)) >
+       (work_area.x + work_area.width)) ||
+      ((cascade_y + (window->frame ? window->frame->rect.height : window->rect.height)) >
+       (work_area.y + work_area.height)))
+    {
+      /* FIXME it would be better to try to start a new cascade to the
+       * right of the last one, probably. Instead of just falling back
+       * to "origin"
+       */
+      cascade_x = MAX (0, work_area.x);
+      cascade_y = MAX (0, work_area.y);
+    }
+  
   /* Convert coords to position of window, not position of frame. */
   if (fgeom == NULL)
     {
@@ -176,6 +200,201 @@ find_next_cascade (MetaWindow *window,
     }
 }
 
+static int
+intcmp (const void* a, const void* b)
+{
+  const int *ai = a;
+  const int *bi = b;
+
+  if (*ai < *bi)
+    return -1;
+  else if (*ai > *bi)
+    return 1;
+  else
+    return 0;
+}
+
+static void
+window_get_edges (MetaWindow *w,
+                  int        *left,
+                  int        *right,
+                  int        *top,
+                  int        *bottom)
+{
+  int left_edge;
+  int right_edge;
+  int top_edge;
+  int bottom_edge;
+  MetaRectangle rect;
+
+  meta_window_get_outer_rect (w, &rect);
+  
+  left_edge = rect.x;
+  right_edge = rect.x + rect.width;
+  top_edge = rect.y;
+  bottom_edge = rect.y + rect.height;
+  
+  if (left)
+    *left = left_edge;
+  if (right)
+    *right = right_edge;
+  if (top)
+    *top = top_edge;
+  if (bottom)
+    *bottom = bottom_edge;
+}
+
+static gboolean
+rectangle_overlaps_some_window (MetaRectangle *rect,
+                                GList         *windows)
+{
+  GList *tmp;
+  MetaRectangle dest;
+  
+  tmp = windows;
+  while (tmp != NULL)
+    {
+      MetaWindow *other = tmp->data;
+      MetaRectangle other_rect;      
+
+      switch (other->type)
+        {
+        case META_WINDOW_DOCK:
+        case META_WINDOW_SPLASHSCREEN:
+        case META_WINDOW_DESKTOP:
+        case META_WINDOW_DIALOG:
+        case META_WINDOW_MODAL_DIALOG:
+          break;
+
+        case META_WINDOW_NORMAL:
+        case META_WINDOW_UTILITY:
+        case META_WINDOW_TOOLBAR:
+        case META_WINDOW_MENU:
+          meta_window_get_outer_rect (other, &other_rect);
+          
+          if (meta_rectangle_intersect (rect, &other_rect, &dest))
+            return TRUE;
+          break;
+        }
+      
+      tmp = tmp->next;
+    }
+
+  return FALSE;
+}
+
+static gint
+leftmost_cmp (gconstpointer a, gconstpointer b)
+{
+  MetaWindow *aw = (gpointer) a;
+  MetaWindow *bw = (gpointer) b;
+  int ax, bx;
+
+  /* we're interested in the frame position for cascading,
+   * not meta_window_get_position()
+   */
+  if (aw->frame)
+    ax = aw->frame->rect.x;
+  else
+    ax = aw->rect.x;
+
+  if (bw->frame)
+    bx = bw->frame->rect.x;
+  else
+    bx = bw->rect.x;
+
+  if (ax < bx)
+    return -1;
+  else if (ax > bx)
+    return 1;
+  else
+    return 0;
+}
+
+static gint
+topmost_cmp (gconstpointer a, gconstpointer b)
+{
+  MetaWindow *aw = (gpointer) a;
+  MetaWindow *bw = (gpointer) b;
+  int ay, by;
+
+  /* we're interested in the frame position for cascading,
+   * not meta_window_get_position()
+   */
+  if (aw->frame)
+    ay = aw->frame->rect.y;
+  else
+    ay = aw->rect.y;
+
+  if (bw->frame)
+    by = bw->frame->rect.y;
+  else
+    by = bw->rect.y;
+
+  if (ay < by)
+    return -1;
+  else if (ay > by)
+    return 1;
+  else
+    return 0;
+}
+
+static gboolean
+fit_rect_in_xinerama (MetaScreen    *screen,
+                      MetaRectangle *rect)
+{
+  int i;
+  int best_index;
+  int best_overlap;
+  const MetaXineramaScreenInfo *xsi;
+  
+  /* Find xinerama with best fit, then
+   * shift rect to be entirely within it.
+   */
+  best_overlap = -1;
+  best_index = -1;
+  
+  i = 0;
+  while (i < screen->n_xinerama_infos)
+    {
+      MetaRectangle xinerama_rect;
+      MetaRectangle intersect;
+      int overlap;
+
+      xsi = &screen->xinerama_infos[i];
+      
+      xinerama_rect.x = xsi->x_origin;
+      xinerama_rect.y = xsi->y_origin;
+      xinerama_rect.width = xsi->width;
+      xinerama_rect.height = xsi->height;
+
+      if (meta_rectangle_intersect (rect, &xinerama_rect, &intersect))
+        overlap = intersect.width * intersect.height;
+      else
+        overlap = 0;
+
+      if (overlap > best_overlap)
+        best_index = i; 
+      
+      ++i;
+    }
+
+  /* some overlap had to be better than -1 */
+  g_assert (best_index >= 0);
+
+  xsi = &screen->xinerama_infos[best_index];
+  
+  if (rect->x < xsi->x_origin)
+    rect->x = xsi->x_origin;
+  if (rect->y < xsi->y_origin)
+    rect->y = xsi->y_origin;
+
+  /* Now return whether we are entirely within the xinerama screen */
+  return
+    ((rect->x + rect->width) < (xsi->x_origin + xsi->width)) &&
+    ((rect->y + rect->height) < (xsi->y_origin + xsi->height));
+}
+
 /* Find the leftmost, then topmost, empty area on the workspace
  * that can contain the new window.
  *
@@ -194,18 +413,156 @@ find_first_fit (MetaWindow *window,
                 int        *new_x,
                 int        *new_y)
 {
-  /* FIXME */
+  /* This algorithm is limited - it just brute-force tries
+   * to fit the window in a small number of locations that are aligned
+   * with existing windows. It tries to place the window on
+   * the bottom of each existing window, and then to the right
+   * of each existing window, aligned with the left/top of the
+   * existing window in each of those cases.
+   */
+  
+  int retval;
+  GList *sorted;
+  GList *tmp;
+  MetaRectangle rect;
+  
+  retval = FALSE;
+
+  rect.width = window->rect.width;
+  rect.height = window->rect.height;
+  
+  if (fgeom)
+    {
+      rect.width += fgeom->left_width + fgeom->right_width;
+      rect.height += fgeom->top_height + fgeom->bottom_height;
+    }
+  
+  sorted = g_list_copy (windows);
+
+  /* Below each window */
+  sorted = g_list_sort (sorted, leftmost_cmp);
+  sorted = g_list_sort (sorted, topmost_cmp);  
+  
+  tmp = sorted;
+  while (tmp != NULL)
+    {
+      MetaWindow *w = tmp->data;
+      MetaRectangle outer_rect;
+      
+      meta_window_get_outer_rect (w, &outer_rect);
+
+      rect.x = outer_rect.x;
+      rect.y = outer_rect.y + outer_rect.height;
+      
+      if (fit_rect_in_xinerama (window->screen, &rect) &&
+          !rectangle_overlaps_some_window (&rect, sorted))
+        {
+          *new_x = rect.x;
+          *new_y = rect.y;
+          if (fgeom)
+            {
+              *new_x += fgeom->left_width;
+              *new_y += fgeom->top_height;
+            }
+
+          retval = TRUE;
+
+          goto out;
+        }      
+      
+      tmp = tmp->next;
+    }
+
+  /* To the right of each window */
+  sorted = g_list_sort (sorted, topmost_cmp);
+  sorted = g_list_sort (sorted, leftmost_cmp);
+  
+  tmp = sorted;
+  while (tmp != NULL)
+    {
+      MetaWindow *w = tmp->data;
+      MetaRectangle outer_rect;
+      
+      meta_window_get_outer_rect (w, &outer_rect);
+
+      rect.x = outer_rect.x + outer_rect.width;
+      rect.y = outer_rect.y;
+      
+      if (fit_rect_in_xinerama (window->screen, &rect) &&
+          !rectangle_overlaps_some_window (&rect, sorted))
+        {
+          *new_x = rect.x;
+          *new_y = rect.y;
+          if (fgeom)
+            {
+              *new_x += fgeom->left_width;
+              *new_y += fgeom->top_height;
+            }
+
+          retval = TRUE;
+
+          goto out;
+        }      
+      
+      tmp = tmp->next;
+    }  
+  
+ out:
+  g_list_free (sorted);
+  return retval;
+}
+
+static void
+constrain_placement (MetaWindow        *window,
+                     MetaFrameGeometry *fgeom,
+                     int                x,
+                     int                y,
+                     int               *new_x,
+                     int               *new_y)
+{
+  /* The purpose of this function is to apply constraints that are not
+   * covered by window.c:constrain_position(), but should apply
+   * whenever we are _placing_ a window regardless of placement algorithm.
+   */
+  MetaRectangle work_area;  
+  int nw_x, nw_y;
+
+  /* FIXME this is bogus because we get the current xinerama
+   * for the window based on its position, but we haven't
+   * placed it yet.
+   */
+  meta_window_get_work_area (window, TRUE, &work_area);
+
+  nw_x = work_area.x;
+  nw_y = work_area.y;
+  if (window->frame)
+    {
+      nw_x += fgeom->left_width;
+      nw_y += fgeom->top_height;
+    }
+
+  /* Keep window from going off left edge, though we don't have
+   * this constraint once the window has been placed.
+   */
+  if (x < nw_x)
+    x = nw_x;
+  if (y < nw_y)
+    y = nw_y;
+
+  *new_x = x;
+  *new_y = y;
 }
 
 void
-meta_window_place (MetaWindow *window,
+meta_window_place (MetaWindow        *window,
                    MetaFrameGeometry *fgeom,
-                   int         x,
-                   int         y,
-                   int        *new_x,
-                   int        *new_y)
+                   int                x,
+                   int                y,
+                   int               *new_x,
+                   int               *new_y)
 {
   GList *windows;
+  const MetaXineramaScreenInfo *xi;
   
   /* frame member variables should NEVER be used in here, only
    * MetaFrameGeometry. But remember fgeom == NULL
@@ -214,13 +571,86 @@ meta_window_place (MetaWindow *window,
    * placement coordinates.
    */
   
-  meta_topic (META_DEBUG_PLACEMENT, "Placing window %s\n", window->desc);      
+  meta_topic (META_DEBUG_PLACEMENT, "Placing window %s\n", window->desc);
 
-  /* FIXME copying Mac, when placing a dialog
-   * put it at 1/5 down and horizontally centered
-   */
+  windows = NULL;
   
-  if (window->xtransient_for != None)
+  switch (window->type)
+    {
+      /* Run placement algorithm on these. */
+    case META_WINDOW_NORMAL:
+    case META_WINDOW_DIALOG:
+    case META_WINDOW_MODAL_DIALOG:
+    case META_WINDOW_SPLASHSCREEN:
+      break;
+          
+      /* Assume the app knows best how to place these, no placement
+       * algorithm ever (other than "leave them as-is")
+       */
+    case META_WINDOW_DESKTOP:
+    case META_WINDOW_DOCK:
+    case META_WINDOW_TOOLBAR:
+    case META_WINDOW_MENU:
+    case META_WINDOW_UTILITY:
+      goto done_no_constraints;
+      break;
+    }
+  
+  if (meta_prefs_get_disable_workarounds ())
+    {
+      switch (window->type)
+        {
+          /* Only accept USPosition on normal windows because the app is full
+           * of shit claiming the user set -geometry for a dialog or dock
+           */
+        case META_WINDOW_NORMAL:
+          if (window->size_hints.flags & USPosition)
+            {
+              /* don't constrain with placement algorithm */
+              meta_topic (META_DEBUG_PLACEMENT,
+                          "Honoring USPosition for %s instead of using placement algorithm\n", window->desc);
+
+              goto done;
+            }
+          break;
+
+          /* Ignore even USPosition on dialogs, splashscreen */
+        case META_WINDOW_DIALOG:
+        case META_WINDOW_MODAL_DIALOG:
+        case META_WINDOW_SPLASHSCREEN:
+          break;
+          
+          /* Assume the app knows best how to place these. */
+        case META_WINDOW_DESKTOP:
+        case META_WINDOW_DOCK:
+        case META_WINDOW_TOOLBAR:
+        case META_WINDOW_MENU:
+        case META_WINDOW_UTILITY:
+          if (window->size_hints.flags & PPosition)
+            {
+              meta_topic (META_DEBUG_PLACEMENT,
+                          "Not placing non-normal non-dialog window with PPosition set\n");
+              goto done_no_constraints;
+            }
+          break;
+        }
+    }
+  else
+    {
+      /* workarounds enabled */
+      
+      if ((window->size_hints.flags & PPosition) ||
+          (window->size_hints.flags & USPosition))
+        {
+          meta_topic (META_DEBUG_PLACEMENT,
+                      "Not placing window with PPosition or USPosition set\n");
+          goto done_no_constraints;
+        }
+    }
+  
+  if ((window->type == META_WINDOW_DIALOG ||
+       window->type == META_WINDOW_MODAL_DIALOG) &&
+      window->xtransient_for != None)
     {
       /* Center horizontally, at top of parent vertically */
 
@@ -242,6 +672,14 @@ meta_window_place (MetaWindow *window,
           /* center of child over center of parent */
           x -= window->rect.width / 2;
 
+          /* put child down 1/5 or so from the top of parent, unless
+           * it makes us have more of parent showing above child than
+           * below
+           */
+          if (window->rect.height <= (parent->rect.height - (parent->rect.height / 5) * 2))
+            y += parent->rect.height / 5;
+
+          /* put top of child's frame, not top of child's client */
           if (fgeom)
             y += fgeom->top_height;
 
@@ -252,6 +690,10 @@ meta_window_place (MetaWindow *window,
         }
     }
 
+  /* FIXME UTILITY with transient set should be stacked up
+   * on the sides of the parent window or something.
+   */
+  
   if (window->type == META_WINDOW_DIALOG ||
       window->type == META_WINDOW_MODAL_DIALOG ||
       window->type == META_WINDOW_SPLASHSCREEN)
@@ -259,15 +701,20 @@ meta_window_place (MetaWindow *window,
       /* Center on screen */
       int w, h;
 
-      /* I think whole screen will look nicer than workarea */
-      w = window->screen->width;
-      h = window->screen->height;
+      /* Warning, this function is a round trip! */
+      xi = meta_screen_get_current_xinerama (window->screen);
+
+      w = xi->width;
+      h = xi->height;
 
       x = (w - window->rect.width) / 2;
       y = (h - window->rect.height) / 2;
 
-      meta_topic (META_DEBUG_PLACEMENT, "Centered window %s on screen\n",
-                  window->desc);
+      x += xi->x_origin;
+      y += xi->y_origin;
+      
+      meta_topic (META_DEBUG_PLACEMENT, "Centered window %s on screen %d xinerama %d\n",
+                  window->desc, window->screen->number, xi->number);
 
       goto done;
     }
@@ -276,7 +723,6 @@ meta_window_place (MetaWindow *window,
    * as placed window, may be shaded - if shaded we pretend it isn't
    * for placement purposes)
    */
-  windows = NULL;
   {
     GSList *all_windows;
     GSList *tmp;
@@ -296,17 +742,25 @@ meta_window_place (MetaWindow *window,
         tmp = tmp->next;
       }
   }
+
+  /* Warning, this is a round trip! */
+  xi = meta_screen_get_current_xinerama (window->screen);
   
   /* "Origin" placement algorithm */
-  x = 0;
-  y = 0;
+  x = xi->x_origin;
+  y = xi->y_origin;
 
-  /* Cascade */
+  if (find_first_fit (window, fgeom, windows, x, y, &x, &y))
+    goto done;
+  
   find_next_cascade (window, fgeom, windows, x, y, &x, &y);
-
-  g_list_free (windows);
   
  done:
+  g_list_free (windows);
+  
+  constrain_placement (window, fgeom, x, y, &x, &y);
+
+ done_no_constraints:
   *new_x = x;
   *new_y = y;
 }
@@ -356,50 +810,6 @@ get_windows_on_same_workspace (MetaWindow *window,
   return windows;
 }
 
-static void
-window_get_edges (MetaWindow *w,
-                  int        *left,
-                  int        *right,
-                  int        *top,
-                  int        *bottom)
-{
-  int left_edge;
-  int right_edge;
-  int top_edge;
-  int bottom_edge;
-  MetaRectangle rect;
-
-  meta_window_get_outer_rect (w, &rect);
-  
-  left_edge = rect.x;
-  right_edge = rect.x + rect.width;
-  top_edge = rect.y;
-  bottom_edge = rect.y + rect.height;
-  
-  if (left)
-    *left = left_edge;
-  if (right)
-    *right = right_edge;
-  if (top)
-    *top = top_edge;
-  if (bottom)
-    *bottom = bottom_edge;
-}
-
-static int
-intcmp (const void* a, const void* b)
-{
-  const int *ai = a;
-  const int *bi = b;
-
-  if (*ai < *bi)
-    return -1;
-  else if (*ai > *bi)
-    return 1;
-  else
-    return 0;
-}
-
 static gboolean
 rects_overlap_vertically (const MetaRectangle *a,
                           const MetaRectangle *b)
@@ -436,7 +846,7 @@ get_vertical_edges (MetaWindow *window,
   GSList *tmp;
   int n_windows;
   int *edges;
-  int i;
+  int i, j;
   int n_edges;
   MetaRectangle rect;
   MetaRectangle work_area;
@@ -444,11 +854,11 @@ get_vertical_edges (MetaWindow *window,
   windows = get_windows_on_same_workspace (window, &n_windows);
 
   i = 0;
-  n_edges = n_windows * 2 + 4; /* 4 = workspace/screen edges */
+  n_edges = n_windows * 2 + 4 + window->screen->n_xinerama_infos - 1; /* 4 = workspace/screen edges */
   edges = g_new (int, n_edges);
 
   /* workspace/screen edges */
-  meta_window_get_work_area (window, &work_area);
+  meta_window_get_work_area (window, FALSE, &work_area);
 
   edges[i] = work_area.x;
   ++i;
@@ -463,6 +873,13 @@ get_vertical_edges (MetaWindow *window,
 
   g_assert (i == 4);
 
+  /* Now get the xinerama screen edges */
+  for (j = 0; j < window->screen->n_xinerama_infos - 1; j++) {
+    edges[i] = window->screen->xinerama_infos[j].x_origin +
+      window->screen->xinerama_infos[j].width;
+    ++i;
+  }
+  
   meta_window_get_outer_rect (window, &rect);
   
   /* get window edges */
@@ -502,7 +919,7 @@ get_horizontal_edges (MetaWindow *window,
   GSList *tmp;
   int n_windows;
   int *edges;
-  int i;
+  int i, j;
   int n_edges;
   MetaRectangle rect;
   MetaRectangle work_area;
@@ -510,11 +927,11 @@ get_horizontal_edges (MetaWindow *window,
   windows = get_windows_on_same_workspace (window, &n_windows);
 
   i = 0;
-  n_edges = n_windows * 2 + 4; /* 4 = workspace/screen edges */
+  n_edges = n_windows * 2 + 4 + window->screen->n_xinerama_infos - 1; /* 4 = workspace/screen edges */
   edges = g_new (int, n_edges);
 
   /* workspace/screen edges */
-  meta_window_get_work_area (window, &work_area);
+  meta_window_get_work_area (window, FALSE, &work_area);
   
   edges[i] = work_area.y;
   ++i;
@@ -528,6 +945,13 @@ get_horizontal_edges (MetaWindow *window,
   ++i;
 
   g_assert (i == 4);
+  
+  /* Now get the xinerama screen edges */
+  for (j = 0; j < window->screen->n_xinerama_infos - 1; j++) {
+    edges[i] = window->screen->xinerama_infos[j].y_origin +
+      window->screen->xinerama_infos[j].height;
+    ++i;
+  }
 
   meta_window_get_outer_rect (window, &rect);
   
