@@ -26,6 +26,17 @@
 #include <X11/Xutil.h>
 #include "ui.h"
 
+typedef struct _MetaXineramaScreenInfo MetaXineramaScreenInfo;
+
+struct _MetaXineramaScreenInfo
+{
+  int number;
+  int x_origin;
+  int y_origin;
+  int width;
+  int height;
+};
+
 typedef void (* MetaScreenWindowFunc) (MetaScreen *screen, MetaWindow *window,
                                        gpointer user_data);
 
@@ -48,6 +59,27 @@ struct _MetaScreen
   MetaStack *stack;
 
   MetaCursor current_cursor;
+
+  Window wm_sn_selection_window;
+  Atom wm_sn_atom;
+  Time wm_sn_timestamp;
+  
+  MetaXineramaScreenInfo *xinerama_infos;
+  int n_xinerama_infos;
+
+  /* Cache the current Xinerama */
+  int last_xinerama_index;
+  
+  guint work_area_idle;
+
+  int rows_of_workspaces;
+  int columns_of_workspaces;
+  guint vertical_workspaces : 1;
+  
+  guint keys_grabbed : 1;
+  guint all_keys_grabbed : 1;
+
+  int closing;
 };
 
 MetaScreen*   meta_screen_new                 (MetaDisplay                *display,
@@ -69,11 +101,25 @@ void          meta_screen_set_cursor          (MetaScreen                 *scree
 void          meta_screen_ensure_tab_popup    (MetaScreen                 *screen,
                                                MetaTabList                 type);
 
+void          meta_screen_ensure_workspace_popup (MetaScreen *screen);
+
 void          meta_screen_focus_top_window    (MetaScreen                 *screen,
                                                MetaWindow                 *not_this_one);
 
+const MetaXineramaScreenInfo* meta_screen_get_current_xinerama (MetaScreen *screen);
+const MetaXineramaScreenInfo* meta_screen_get_xinerama_for_window (MetaScreen *screen,
+								   MetaWindow *window);
+
+void          meta_screen_update_workspace_layout (MetaScreen             *screen);
+void          meta_screen_update_workspace_names  (MetaScreen             *screen);
+void          meta_screen_queue_workarea_recalc   (MetaScreen             *screen);
+
+Window meta_create_offscreen_window (Display *xdisplay,
+                                     Window   parent);
+
+void meta_screen_calc_workspace_layout (MetaScreen *screen,
+                                        int         num_workspaces,
+                                        int        *r,
+                                        int        *c);
+
 #endif
-
-
-
-
