@@ -34,6 +34,12 @@ meta_session_init (const char *client_id,
   meta_topic (META_DEBUG_SM, "Compiled without session management support\n");
 }
 
+void
+meta_session_shutdown (void)
+{
+  /* nothing */
+}
+
 const MetaWindowSessionInfo*
 meta_window_lookup_saved_state (MetaWindow *window)
 {
@@ -73,6 +79,7 @@ static char*       load_state         (const char *previous_save_file);
 static void        regenerate_save_file (void);
 static const char* full_save_file       (void);
 static const char* base_save_file       (void);
+static void        warn_about_lame_clients_and_finish_interact (gboolean shutdown);
 
 /* This is called when data is available on an ICE connection.  */
 static gboolean
@@ -181,6 +188,9 @@ typedef enum
   STATE_SAVING_PHASE_1,
   STATE_WAITING_FOR_PHASE_2,
   STATE_SAVING_PHASE_2,
+  STATE_WAITING_FOR_INTERACT,
+  STATE_DONE_WITH_INTERACT,
+  STATE_SKIPPING_GLOBAL_SAVE,
   STATE_FROZEN,
   STATE_REGISTERING
 } ClientState;
@@ -206,6 +216,7 @@ static void set_clone_restart_commands  (void);
 static char *client_id = NULL;
 static gpointer session_connection = NULL;
 static ClientState current_state = STATE_DISCONNECTED;
+static gboolean interaction_allowed = FALSE;
 
 void
 meta_session_init (const char *previous_client_id,
@@ -234,12 +245,6 @@ meta_session_init (const char *previous_client_id,
   
   mask = SmcSaveYourselfProcMask | SmcDieProcMask |
     SmcSaveCompleteProcMask | SmcShutdownCancelledProcMask;
-
-#if 0
-  /* note this isn't in the mask above */
-  callbacks.interact.callback = interact_callback;
-  callbacks.interact.client_data = NULL;
-#endif
   
   callbacks.save_yourself.callback = save_yourself_callback;
   callbacks.save_yourself.client_data = NULL;
@@ -266,7 +271,7 @@ meta_session_init (const char *previous_client_id,
   
   if (session_connection == NULL)
     {
-      meta_warning ("Failed to open connection to session manager: %s\n", buf);
+      meta_warning (_("Failed to a open connection to a session manager, so window positions will not be saved: %s\n"), buf);
 
       goto out;
     }
@@ -286,7 +291,7 @@ meta_session_init (const char *previous_client_id,
     SmProp prop1, prop2, prop3, prop4, prop5, prop6, *props[6];
     SmPropValue prop1val, prop2val, prop3val, prop4val, prop5val, prop6val;
     char pid[32];
-    char hint = SmRestartIfRunning;
+    char hint = SmRestartImmediately;
     char priority = 20; /* low to run before other apps */
     
     prop1.name = SmProgram;
@@ -350,6 +355,31 @@ meta_session_init (const char *previous_client_id,
   g_free (saved_client_id);
 }
 
+void
+meta_session_shutdown (void)
+{
+  /* Change our restart mode to IfRunning */
+  
+  SmProp prop1;
+  SmPropValue prop1val;
+  SmProp *props[1];
+  char hint = SmRestartIfRunning;
+
+  if (session_connection == NULL)
+    return;
+  
+  prop1.name = SmRestartStyleHint;
+  prop1.type = SmCARD8;
+  prop1.num_vals = 1;
+  prop1.vals = &prop1val;
+  prop1val.value = &hint;
+  prop1val.length = 1;
+    
+  props[0] = &prop1;
+  
+  SmcSetProperties (session_connection, 1, props);
+}
+
 static void
 disconnect (void)
 {
@@ -362,6 +392,10 @@ static void
 save_yourself_possibly_done (gboolean shutdown,
                              gboolean successful)
 {
+  meta_topic (META_DEBUG_SM,
+              "save possibly done shutdown = %d success = %d\n",
+              shutdown, successful);
+  
   if (current_state == STATE_SAVING_PHASE_1)
     {
       Status status;
@@ -372,11 +406,38 @@ save_yourself_possibly_done (gboolean shutdown,
 
       if (status)
         current_state = STATE_WAITING_FOR_PHASE_2;
+
+      meta_topic (META_DEBUG_SM,
+                  "Requested phase 2, status = %d\n", status);
     }
 
-  if (current_state == STATE_SAVING_PHASE_1 ||
-      current_state == STATE_SAVING_PHASE_2)
+  if (current_state == STATE_SAVING_PHASE_2 &&
+      interaction_allowed)
     {
+      Status status;
+
+      status = SmcInteractRequest (session_connection,
+                                   /* ignore this feature of the protocol by always
+                                    * claiming normal
+                                    */
+                                   SmDialogNormal,
+                                   interact_callback,
+                                   GINT_TO_POINTER (shutdown));
+
+      if (status)
+        current_state = STATE_WAITING_FOR_INTERACT;
+
+      meta_topic (META_DEBUG_SM,
+                  "Requested interact, status = %d\n", status);
+    }
+  
+  if (current_state == STATE_SAVING_PHASE_1 ||
+      current_state == STATE_SAVING_PHASE_2 ||
+      current_state == STATE_DONE_WITH_INTERACT ||
+      current_state == STATE_SKIPPING_GLOBAL_SAVE)
+    {
+      meta_topic (META_DEBUG_SM, "Sending SaveYourselfDone\n");
+      
       SmcSaveYourselfDone (session_connection,
                            successful);
       
@@ -387,12 +448,13 @@ save_yourself_possibly_done (gboolean shutdown,
     }
 }
 
-
 static void 
 save_phase_2_callback (SmcConn smc_conn, SmPointer client_data)
 {
   gboolean shutdown;
 
+  meta_topic (META_DEBUG_SM, "Phase 2 save");
+  
   shutdown = GPOINTER_TO_INT (client_data);
   
   current_state = STATE_SAVING_PHASE_2;
@@ -412,6 +474,8 @@ save_yourself_callback (SmcConn   smc_conn,
 {
   gboolean successful;
 
+  meta_topic (META_DEBUG_SM, "SaveYourself received");
+  
   successful = TRUE;
   
   /* The first SaveYourself after registering for the first time
@@ -435,6 +499,22 @@ save_yourself_callback (SmcConn   smc_conn,
     }
 #endif
 
+  /* ignore Global style saves
+   * 
+   * This interpretaion of the Local/Global/Both styles
+   * was discussed extensively on the xdg-list. See:
+   *
+   * https://listman.redhat.com/pipermail/xdg-list/2002-July/000615.html
+   */
+  if (save_style == SmSaveGlobal)
+    {
+      current_state = STATE_SKIPPING_GLOBAL_SAVE;
+      save_yourself_possibly_done (shutdown, successful);
+      return;
+    }
+
+  interaction_allowed = interact_style != SmInteractStyleNone;
+  
   current_state = STATE_SAVING_PHASE_1;
 
   regenerate_save_file ();
@@ -457,18 +537,35 @@ static void
 save_complete_callback (SmcConn smc_conn, SmPointer client_data)
 {
   /* nothing */
+  meta_topic (META_DEBUG_SM, "SaveComplete received\n");
 }
 
 static void
 shutdown_cancelled_callback (SmcConn smc_conn, SmPointer client_data)
 {
-  /* nothing */
+  meta_topic (META_DEBUG_SM, "Shutdown cancelled received\n");
+  
+  if (session_connection != NULL &&
+      (current_state != STATE_IDLE && current_state != STATE_FROZEN))
+    {
+      SmcSaveYourselfDone (session_connection, True);
+      current_state = STATE_IDLE;
+    }
 }
 
 static void 
 interact_callback (SmcConn smc_conn, SmPointer client_data)
 {
   /* nothing */
+  gboolean shutdown;
+
+  meta_topic (META_DEBUG_SM, "Interaction permission received\n");
+  
+  shutdown = GPOINTER_TO_INT (client_data);
+
+  current_state = STATE_DONE_WITH_INTERACT;
+
+  warn_about_lame_clients_and_finish_interact (shutdown);
 }
 
 static void
@@ -741,6 +838,24 @@ decode_text_from_utf8 (const char *text)
   return g_string_free (str, FALSE);
 }
 
+static int
+stack_cmp (const void *a,
+           const void *b)
+{
+  MetaWindow *aw = (void*) a;
+  MetaWindow *bw = (void*) b;
+
+  if (aw->screen == bw->screen)
+    return meta_stack_windows_cmp (aw->screen->stack, aw, bw);
+  /* Then assume screens are stacked by number */
+  else if (aw->screen->number < bw->screen->number)
+    return -1;
+  else if (aw->screen->number > bw->screen->number)
+    return 1;
+  else
+    return 0; /* not reached in theory, if windows on same display */
+}
+
 static void
 save_state (void)
 {
@@ -787,10 +902,10 @@ save_state (void)
 
   /* The file format is:
    * <metacity_session id="foo">
-   *   <window id="bar" class="XTerm" name="xterm" title="/foo/bar" role="blah" type="normal">
+   *   <window id="bar" class="XTerm" name="xterm" title="/foo/bar" role="blah" type="normal" stacking="5">
    *     <workspace index="2"/>
    *     <workspace index="4"/>
-   *     <sticky/>
+   *     <sticky/> <minimized/> <maximized/>
    *     <geometry x="100" y="100" width="200" height="200" gravity="northwest"/>
    *   </window>
    * </metacity_session>
@@ -810,8 +925,12 @@ save_state (void)
     {
       GSList *windows;
       GSList *tmp;
-
+      int stack_position;
+      
       windows = meta_display_list_windows (display_iter->data);
+      windows = g_slist_sort (windows, stack_cmp);
+
+      stack_position = 0;
       tmp = windows;
       while (tmp != NULL)
         {
@@ -843,13 +962,14 @@ save_state (void)
                           window->desc, window->sm_client_id);
 
               fprintf (outfile,
-                       "  <window id=\"%s\" class=\"%s\" name=\"%s\" title=\"%s\" role=\"%s\" type=\"%s\">\n",
+                       "  <window id=\"%s\" class=\"%s\" name=\"%s\" title=\"%s\" role=\"%s\" type=\"%s\" stacking=\"%d\">\n",
                        sm_client_id,
                        res_class ? res_class : "",
                        res_name ? res_name : "",
                        window->title ? window->title : "",
                        role ? role : "",
-                       window_type_to_string (window->type));
+                       window_type_to_string (window->type),
+                       stack_position);
 
               g_free (sm_client_id);
               g_free (res_class);
@@ -860,6 +980,14 @@ save_state (void)
               if (window->on_all_workspaces)
                 fputs ("    <sticky/>\n", outfile);
 
+              /* Minimized */
+              if (window->minimized)
+                fputs ("    <minimized/>\n", outfile);
+
+              /* Maximized */
+              if (window->maximized)
+                fputs ("    <maximized/>\n", outfile);
+              
               /* Workspaces we're on */
               {
                 GList *w;
@@ -895,6 +1023,7 @@ save_state (void)
             }
           
           tmp = tmp->next;
+          ++stack_position;
         }
       
       g_slist_free (windows);
@@ -931,6 +1060,8 @@ typedef enum
   WINDOW_TAG_NONE,
   WINDOW_TAG_DESKTOP,
   WINDOW_TAG_STICKY,
+  WINDOW_TAG_MINIMIZED,
+  WINDOW_TAG_MAXIMIZED,
   WINDOW_TAG_GEOMETRY
 } WindowTag;
 
@@ -976,7 +1107,7 @@ load_state (const char *previous_save_file)
   GError *error;
   ParseData parse_data;
   char *text;
-  int length;
+  gsize length;
   char *session_file;
 
   session_file = g_strconcat (g_get_home_dir (),
@@ -1149,6 +1280,14 @@ start_element_handler  (GMarkupParseContext *context,
               if (*val)
                 pd->info->type = window_type_from_string (val);
             }
+          else if (strcmp (name, "stacking") == 0)
+            {
+              if (*val)
+                {
+                  pd->info->stack_position = atoi (val);
+                  pd->info->stack_position_set = TRUE;
+                }
+            }
           else
             {
               g_set_error (error,
@@ -1201,6 +1340,16 @@ start_element_handler  (GMarkupParseContext *context,
       pd->info->on_all_workspaces = TRUE;
       pd->info->on_all_workspaces_set = TRUE;
     }
+  else if (strcmp (element_name, "minimized") == 0)
+    {
+      pd->info->minimized = TRUE;
+      pd->info->minimized_set = TRUE;
+    }
+  else if (strcmp (element_name, "maximized") == 0)
+    {
+      pd->info->maximized = TRUE;
+      pd->info->maximized_set = TRUE;
+    }  
   else if (strcmp (element_name, "geometry") == 0)
     {
       int i;
@@ -1546,4 +1695,173 @@ base_save_file (void)
   return relative_save_path;
 }
 
+static int
+windows_cmp_by_title (MetaWindow *a,
+                      MetaWindow *b)
+{
+  return g_utf8_collate (a->title, b->title);
+}
+
+typedef struct
+{
+  int child_pid;
+  int child_pipe;
+  gboolean shutdown;
+} LameClientsDialogData;
+
+static void
+finish_interact (gboolean shutdown)
+{
+  if (current_state == STATE_DONE_WITH_INTERACT) /* paranoia */
+    {
+      SmcInteractDone (session_connection, False /* don't cancel logout */);
+      
+      save_yourself_possibly_done (shutdown, TRUE);
+    }
+}
+
+static gboolean  
+io_from_warning_dialog (GIOChannel   *channel,
+                        GIOCondition  condition,
+                        gpointer      data)
+{
+  LameClientsDialogData *d;
+
+  d = data;
+  
+  meta_topic (META_DEBUG_PING,
+              "IO handler from lame clients dialog, condition = %x\n",
+              condition);
+  
+  if (condition & (G_IO_HUP | G_IO_NVAL | G_IO_ERR))
+    {
+      finish_interact (d->shutdown);
+
+      /* Remove the callback, freeing data */
+      return FALSE; 
+    }
+
+  /* Keep callback installed */
+  return TRUE;
+}
+
+static void
+warn_about_lame_clients_and_finish_interact (gboolean shutdown)
+{
+  GSList *displays;
+  GSList *display_iter;  
+  GSList *lame;
+  char **argv;
+  int i;
+  GSList *tmp;
+  int len;
+  int child_pid;
+  int child_pipe;
+  GError *err;
+  GIOChannel *channel;
+  LameClientsDialogData *d;
+  
+  lame = NULL;
+  displays = meta_displays_list ();
+  display_iter = displays;
+  while (display_iter != NULL)
+    {
+      GSList *windows;
+
+      windows = meta_display_list_windows (display_iter->data);
+      tmp = windows;
+      while (tmp != NULL)
+        {
+          MetaWindow *window;
+
+          window = tmp->data;
+
+          /* only complain about normal windows, the others
+           * are kind of dumb to worry about
+           */
+          if (window->sm_client_id == NULL &&
+              window->type == META_WINDOW_NORMAL)
+            lame = g_slist_prepend (lame, window);
+          
+          tmp = tmp->next;
+        }
+      
+      g_slist_free (windows);
+
+      display_iter = display_iter->next;
+    }
+  /* don't need to free displays */
+  displays = NULL;
+
+  if (lame == NULL)
+    {
+      /* No lame apps. */
+      finish_interact (shutdown);
+      return;
+    }
+  
+  lame = g_slist_sort (lame, (GCompareFunc) windows_cmp_by_title);
+
+  len = g_slist_length (lame);
+  len *= 2; /* titles and also classes */
+  len += 1; /* NULL term */
+  len += 2; /* metacity-dialog command and option */
+  
+  argv = g_new0 (char*, len);
+  
+  i = 0;
+
+  argv[i] = METACITY_LIBEXECDIR"/metacity-dialog";
+  ++i;
+  argv[i] = "--warn-about-no-sm-support";
+  ++i;
+  
+  tmp = lame;
+  while (tmp != NULL)
+    {
+      MetaWindow *w = tmp->data;
+
+      argv[i] = w->title;
+      ++i;
+      argv[i] = w->res_class ? w->res_class : "";
+      ++i;
+
+      tmp = tmp->next;
+    }
+
+  child_pipe = -1;
+  child_pid = -1;
+  err = NULL;
+  if (!g_spawn_async_with_pipes ("/",
+                                 argv,
+                                 NULL,
+                                 0,
+                                 NULL, NULL,
+                                 &child_pid,
+                                 NULL,
+                                 &child_pipe,
+                                 NULL,
+                                 &err))
+    {
+      meta_warning (_("Error launching metacity-dialog to warn about apps that don't support session management: %s\n"),
+                    err->message);
+      g_error_free (err);
+    }
+
+  g_free (argv);
+  g_slist_free (lame);
+
+  d = g_new0 (LameClientsDialogData, 1);
+  d->child_pipe = child_pipe;
+  d->child_pid = child_pid;
+  d->shutdown = shutdown;
+  
+  channel = g_io_channel_unix_new (d->child_pipe);
+  g_io_add_watch_full (channel, G_PRIORITY_DEFAULT,
+                       G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
+                       io_from_warning_dialog,
+                       d, g_free);
+  g_io_channel_unref (channel);
+}
+
 #endif /* HAVE_SM */
