@@ -267,7 +267,7 @@ compute_layer (MetaWindow *window)
 
     case META_WINDOW_DOCK:
       /* still experimenting here */
-      window->layer = META_LAYER_NORMAL;
+      window->layer = META_LAYER_DOCK;
       break;
 
     case META_WINDOW_SPLASHSCREEN:
@@ -370,6 +370,105 @@ sort_window_list (GList *list)
   return list;
 }
 
+static void
+raise_window_relative_to_managed_windows (MetaScreen *screen,
+                                          Window      xwindow)
+{
+  /* This function is used to avoid raising a window above popup
+   * menus and other such things.
+   *
+   * FIXME This is sort of an expensive function, should probably
+   * do something to avoid it. One approach would be to reverse
+   * the stacking algorithm to work by placing each window above
+   * the others, and start by lowering a window to the bottom
+   * (instead of the current way, which works by placing each
+   * window below another and starting with a raise)
+   */
+
+  Window ignored1, ignored2;
+  Window *children;
+  int n_children;
+  int i;
+
+  /* Normally XQueryTree() means "must grab server" but here
+   * we don't, since we know we won't manage any new windows
+   * or restack any windows before using the XQueryTree results.
+   */
+  
+  meta_error_trap_push (screen->display);
+  
+  XQueryTree (screen->display->xdisplay,
+              screen->xroot,
+              &ignored1, &ignored2, &children, &n_children);
+
+  if (meta_error_trap_pop (screen->display))
+    {
+      meta_topic (META_DEBUG_STACK,
+                  "Error querying root children to raise window 0x%lx\n",
+                  xwindow);
+      return;
+    }
+
+  /* Children are in order from bottom to top. We want to
+   * find the topmost managed child, then configure
+   * our window to be above it.
+   */
+  i = n_children - 1;
+  while (i >= 0)
+    {
+      if (children[i] == xwindow)
+        {
+          /* Do nothing. This means we're already the topmost managed
+           * window, but it DOES NOT mean we are already just above
+           * the topmost managed window. This is important because if
+           * an override redirect window is up, and we map a new
+           * managed window, the new window is probably above the old
+           * popup by default, and we want to push it below that
+           * popup. So keep looking for a sibling managed window
+           * to be moved below.
+           */
+        }
+      else if (meta_display_lookup_x_window (screen->display,
+                                             children[i]) != NULL)
+        {
+          XWindowChanges changes;
+          
+          /* children[i] is the topmost managed child */
+          meta_topic (META_DEBUG_STACK,
+                      "Moving 0x%lx above topmost managed child window 0x%lx\n",
+                      xwindow, children[i]);
+
+          changes.sibling = children[i];
+          changes.stack_mode = Above;
+
+          meta_error_trap_push (screen->display);
+          XConfigureWindow (screen->display->xdisplay,
+                            xwindow,
+                            CWSibling | CWStackMode,
+                            &changes);
+          meta_error_trap_pop (screen->display);
+
+          break;
+        }
+
+      --i;
+    }
+
+  if (i < 0)
+    {
+      /* No sibling to use, just lower ourselves to the bottom
+       * to be sure we're below any override redirect windows.
+       */
+      meta_error_trap_push (screen->display);
+      XLowerWindow (screen->display->xdisplay,
+                    xwindow);
+      meta_error_trap_pop (screen->display);
+    }
+  
+  if (children)
+    XFree (children);
+}                                         
+
 static void
 meta_stack_sync_to_server (MetaStack *stack)
 {
@@ -696,10 +795,10 @@ meta_stack_sync_to_server (MetaStack *stack)
               /* Move *newp below last_window */
               if (last_window == None)
                 {
-                  meta_topic (META_DEBUG_STACK, "Raising window 0x%lx to the top\n", *newp);
-                  
-                  XRaiseWindow (stack->screen->display->xdisplay,
-                                *newp);
+                  meta_topic (META_DEBUG_STACK, "Using window 0x%lx as topmost (but leaving it in-place)\n", *newp);
+
+                  raise_window_relative_to_managed_windows (stack->screen,
+                                                            *newp);
                 }
               else
                 {
@@ -839,7 +938,8 @@ meta_stack_get_bottom (MetaStack  *stack)
 
 MetaWindow*
 meta_stack_get_above (MetaStack      *stack,
-                      MetaWindow     *window)
+                      MetaWindow     *window,
+                      gboolean        only_within_layer)
 {
   GList *link;
 
@@ -852,13 +952,16 @@ meta_stack_get_above (MetaStack      *stack,
 
   if (link->prev)
     return link->prev->data;
+  else if (only_within_layer)
+    return NULL;
   else
     return find_next_above_layer (stack, window->layer);
 }
 
 MetaWindow*
 meta_stack_get_below (MetaStack      *stack,
-                      MetaWindow     *window)
+                      MetaWindow     *window,
+                      gboolean        only_within_layer)
 {
   GList *link;
 
@@ -871,6 +974,8 @@ meta_stack_get_below (MetaStack      *stack,
 
   if (link->next)
     return link->next->data;
+  else if (only_within_layer)
+    return NULL;
   else
     return find_prev_below_layer (stack, window->layer);
 }                               
@@ -906,13 +1011,14 @@ meta_stack_get_default_focus_window (MetaStack     *stack,
           if (window &&
               window != not_this_one &&
               (window->unmaps_pending == 0) &&
+              !window->minimized &&
               (workspace == NULL ||
                meta_window_visible_on_workspace (window, workspace)))
             {
               if (topmost_dock == NULL &&
                   window->type == META_WINDOW_DOCK)
                 topmost_dock = window;
-              else
+              else if (window->type != META_WINDOW_DOCK)
                 return window;
             }
 
@@ -929,6 +1035,42 @@ meta_stack_get_default_focus_window (MetaStack     *stack,
   return topmost_dock;
 }
 
+GList*
+meta_stack_list_windows (MetaStack     *stack,
+                         MetaWorkspace *workspace)
+{
+  GList *workspace_windows = NULL;
+  int layer = META_LAYER_LAST;  
+
+  --layer;
+  while (layer >= 0)
+    {
+      GList *link;
+
+      g_assert (layer >= 0 && layer < META_LAYER_LAST);
+
+      /* top of this layer is at the front of the list */
+      link = stack->layers[layer];
+      
+      while (link)
+        {
+          MetaWindow *window = link->data;
+
+          if (window && meta_window_visible_on_workspace (window, workspace))
+            {
+              workspace_windows = g_list_prepend (workspace_windows,
+                                                  window);
+            }
+
+          link = link->next;
+        }
+      
+      --layer;
+    }
+
+  return workspace_windows;
+}
+
 int
 meta_stack_windows_cmp  (MetaStack  *stack,
                          MetaWindow *window_a,
