@@ -19,9 +19,14 @@
  * 02111-1307, USA.
  */
 
+#include <config.h>
+
 #include "util.h"
 #include "core.h"
 #include "tabpopup.h"
+#include "workspace.h" /* FIXME should not be included in this file */
+#include "draw-workspace.h"
+#include "frame.h"
 #include <gtk/gtk.h>
 #include <math.h>
 
@@ -32,12 +37,12 @@ typedef struct _TabEntry TabEntry;
 
 struct _TabEntry
 {
-  Window      xwindow;
-  char       *title;
-  GdkPixbuf  *icon;
-  GtkWidget  *widget;
-  GdkRectangle rect;
-  GdkRectangle inner_rect;
+  MetaTabEntryKey  key;
+  char            *title;
+  GdkPixbuf       *icon;
+  GtkWidget       *widget;
+  GdkRectangle     rect;
+  GdkRectangle     inner_rect;
 };
 
 struct _MetaTabPopup
@@ -48,12 +53,17 @@ struct _MetaTabPopup
   GList *entries;
   TabEntry *current_selected_entry;
   GtkWidget *outline_window;
+  gboolean outline;
 };
 
 static GtkWidget* selectable_image_new (GdkPixbuf *pixbuf);
 static void       select_image         (GtkWidget *widget);
 static void       unselect_image       (GtkWidget *widget);
 
+static GtkWidget* selectable_workspace_new (MetaWorkspace *workspace);
+static void       select_workspace         (GtkWidget *widget);
+static void       unselect_workspace       (GtkWidget *widget);
+
 static gboolean
 outline_window_expose (GtkWidget      *widget,
                        GdkEventExpose *event,
@@ -65,7 +75,7 @@ outline_window_expose (GtkWidget      *widget,
   
   popup = data;
 
-  if (popup->current_selected_entry == NULL)
+  if (!popup->outline || popup->current_selected_entry == NULL)
     return FALSE;
 
   te = popup->current_selected_entry;
@@ -84,18 +94,21 @@ outline_window_expose (GtkWidget      *widget,
                       FALSE,
                       te->inner_rect.x - 1, te->inner_rect.y - 1,
                       te->inner_rect.width + 1,
-                      te->inner_rect.height + 1);  
-  
+                      te->inner_rect.height + 1);
+
   return FALSE;
 }
 
 MetaTabPopup*
-meta_ui_tab_popup_new (const MetaTabEntry *entries)
+meta_ui_tab_popup_new (const MetaTabEntry *entries,
+                       int                 screen_number,
+                       int                 entry_count,
+                       int                 width,
+                       gboolean            outline)
 {
   MetaTabPopup *popup;
   int i, left, right, top, bottom;
   GList *tab_entries;
-  int width;
   int height;
   GtkWidget *table;
   GtkWidget *vbox;
@@ -103,10 +116,16 @@ meta_ui_tab_popup_new (const MetaTabEntry *entries)
   GList *tmp;
   GtkWidget *frame;
   int max_label_width;
-  
+
   popup = g_new (MetaTabPopup, 1);
 
   popup->outline_window = gtk_window_new (GTK_WINDOW_POPUP);
+#ifdef HAVE_GTK_MULTIHEAD
+  gtk_window_set_screen (GTK_WINDOW (popup->outline_window),
+			 gdk_display_get_screen (gdk_display_get_default (),
+						 screen_number));
+#endif
+
   gtk_widget_set_app_paintable (popup->outline_window, TRUE);
   gtk_widget_realize (popup->outline_window);
 
@@ -114,6 +133,12 @@ meta_ui_tab_popup_new (const MetaTabEntry *entries)
                     G_CALLBACK (outline_window_expose), popup);
   
   popup->window = gtk_window_new (GTK_WINDOW_POPUP);
+#ifdef HAVE_GTK_MULTIHEAD
+  gtk_window_set_screen (GTK_WINDOW (popup->window),
+			 gdk_display_get_screen (gdk_display_get_default (),
+						 screen_number));
+#endif
+
   gtk_window_set_position (GTK_WINDOW (popup->window),
                            GTK_WIN_POS_CENTER_ALWAYS);
   /* enable resizing, to get never-shrink behavior */
@@ -122,38 +147,40 @@ meta_ui_tab_popup_new (const MetaTabEntry *entries)
   popup->current = NULL;
   popup->entries = NULL;
   popup->current_selected_entry = NULL;
+  popup->outline = outline;
   
   tab_entries = NULL;
-  i = 0;
-  while (entries[i].xwindow != None)
+  for (i = 0; i < entry_count; ++i)
     {
       TabEntry *te;
 
       te = g_new (TabEntry, 1);
-      te->xwindow = entries[i].xwindow;
+      te->key = entries[i].key;
       te->title = g_strdup (entries[i].title);
-      te->icon = entries[i].icon;
-      g_object_ref (G_OBJECT (te->icon));
       te->widget = NULL;
+      te->icon = entries[i].icon;
+      if (te->icon)
+        g_object_ref (G_OBJECT (te->icon));
 
-      te->rect.x = entries[i].x;
-      te->rect.y = entries[i].y;
-      te->rect.width = entries[i].width;
-      te->rect.height = entries[i].height;
+      if (outline)
+        {
+          te->rect.x = entries[i].x;
+          te->rect.y = entries[i].y;
+          te->rect.width = entries[i].width;
+          te->rect.height = entries[i].height;
+
+          te->inner_rect.x = entries[i].inner_x;
+          te->inner_rect.y = entries[i].inner_y;
+          te->inner_rect.width = entries[i].inner_width;
+          te->inner_rect.height = entries[i].inner_height;
+        }
 
-      te->inner_rect.x = entries[i].inner_x;
-      te->inner_rect.y = entries[i].inner_y;
-      te->inner_rect.width = entries[i].inner_width;
-      te->inner_rect.height = entries[i].inner_height;
-      
       tab_entries = g_list_prepend (tab_entries, te);
-      
-      ++i;
     }
 
   popup->entries = g_list_reverse (tab_entries);
 
-  width = 5; /* FIXME */
+  g_assert (width > 0);
   height = i / width;
   if (i % width)
     height += 1;
@@ -199,12 +226,20 @@ meta_ui_tab_popup_new (const MetaTabEntry *entries)
           TabEntry *te;
 
           te = tmp->data;
-          
-          image = selectable_image_new (te->icon);
-          gtk_misc_set_padding (GTK_MISC (image),
-                                INSIDE_SELECT_RECT + OUTSIDE_SELECT_RECT + 1,
-                                INSIDE_SELECT_RECT + OUTSIDE_SELECT_RECT + 1);
-          gtk_misc_set_alignment (GTK_MISC (image), 0.5, 0.5);
+       
+          if (outline)
+            {
+              image = selectable_image_new (te->icon);
+
+              gtk_misc_set_padding (GTK_MISC (image),
+                                    INSIDE_SELECT_RECT + OUTSIDE_SELECT_RECT + 1,
+                                    INSIDE_SELECT_RECT + OUTSIDE_SELECT_RECT + 1);
+              gtk_misc_set_alignment (GTK_MISC (image), 0.5, 0.5);
+            }   
+          else
+            {
+              image = selectable_workspace_new ((MetaWorkspace *) te->key);
+            }
           
           te->widget = image;
 
@@ -250,7 +285,8 @@ free_entry (gpointer data, gpointer user_data)
   te = data;
   
   g_free (te->title);
-  g_object_unref (G_OBJECT (te->icon));
+  if (te->icon)
+    g_object_unref (G_OBJECT (te->icon));
 
   g_free (te);
 }
@@ -280,9 +316,12 @@ meta_ui_tab_popup_set_showing (MetaTabPopup *popup,
     }
   else
     {
-      meta_verbose ("Hiding tab popup window\n");
-      gtk_widget_hide (popup->window);
-      meta_core_increment_event_serial (gdk_display);
+      if (GTK_WIDGET_VISIBLE (popup->window))
+        {
+          meta_verbose ("Hiding tab popup window\n");
+          gtk_widget_hide (popup->window);
+          meta_core_increment_event_serial (gdk_display);
+        }
     }
 }
 
@@ -296,41 +335,53 @@ display_entry (MetaTabPopup *popup,
 
   
   if (popup->current_selected_entry)
-    unselect_image (popup->current_selected_entry->widget);
+  {
+    if (popup->outline)
+      unselect_image (popup->current_selected_entry->widget);
+    else
+      unselect_workspace (popup->current_selected_entry->widget);
+  }
   
   gtk_label_set_text (GTK_LABEL (popup->label), te->title);
-  select_image (te->widget);
-  
-  /* Do stuff behind gtk's back */
-  gdk_window_hide (popup->outline_window->window);
-  meta_core_increment_event_serial (gdk_display);
-  
-  rect = te->rect;
-  rect.x = 0;
-  rect.y = 0;
 
-  gdk_window_move_resize (popup->outline_window->window,
-                          te->rect.x, te->rect.y,
-                          te->rect.width, te->rect.height);
-  
-  gdk_window_set_background (popup->outline_window->window,
-                             &popup->outline_window->style->black);
-  
-  region = gdk_region_rectangle (&rect);
-  inner_region = gdk_region_rectangle (&te->inner_rect);
-  gdk_region_subtract (region, inner_region);
-  gdk_region_destroy (inner_region);
-  
-  gdk_window_shape_combine_region (popup->outline_window->window,
-                                   region,
-                                   0, 0);
+  if (popup->outline)
+    select_image (te->widget);
+  else
+    select_workspace (te->widget);
 
-  gdk_region_destroy (region);
+  if (popup->outline)
+    {
+      /* Do stuff behind gtk's back */
+      gdk_window_hide (popup->outline_window->window);
+      meta_core_increment_event_serial (gdk_display);
   
-  /* This should piss off gtk a bit, but we don't want to raise
-   * above the tab popup
-   */
-  gdk_window_show_unraised (popup->outline_window->window);
+      rect = te->rect;
+      rect.x = 0;
+      rect.y = 0;
+
+      gdk_window_move_resize (popup->outline_window->window,
+                              te->rect.x, te->rect.y,
+                              te->rect.width, te->rect.height);
+  
+      gdk_window_set_background (popup->outline_window->window,
+                                 &popup->outline_window->style->black);
+  
+      region = gdk_region_rectangle (&rect);
+      inner_region = gdk_region_rectangle (&te->inner_rect);
+      gdk_region_subtract (region, inner_region);
+      gdk_region_destroy (inner_region);
+  
+      gdk_window_shape_combine_region (popup->outline_window->window,
+                                       region,
+                                       0, 0);
+
+      gdk_region_destroy (region);
+  
+      /* This should piss off gtk a bit, but we don't want to raise
+       * above the tab popup
+       */
+      gdk_window_show_unraised (popup->outline_window->window);
+    }
 
   /* Must be before we handle an expose for the outline window */
   popup->current_selected_entry = te;
@@ -374,7 +425,7 @@ meta_ui_tab_popup_backward (MetaTabPopup *popup)
     }
 }
 
-Window
+MetaTabEntryKey
 meta_ui_tab_popup_get_selected (MetaTabPopup *popup)
 {
   if (popup->current)
@@ -383,7 +434,7 @@ meta_ui_tab_popup_get_selected (MetaTabPopup *popup)
 
       te = popup->current->data;
 
-      return te->xwindow;
+      return te->key;
     }
   else
     return None;
@@ -391,7 +442,7 @@ meta_ui_tab_popup_get_selected (MetaTabPopup *popup)
 
 void
 meta_ui_tab_popup_select (MetaTabPopup *popup,
-                          Window        xwindow)
+                          MetaTabEntryKey key)
 {
   GList *tmp;
 
@@ -402,7 +453,7 @@ meta_ui_tab_popup_select (MetaTabPopup *popup,
 
       te = tmp->data;
 
-      if (te->xwindow == xwindow)
+      if (te->key == key)
         {
           popup->current = tmp;
           
@@ -529,7 +580,7 @@ meta_select_image_expose_event (GtkWidget      *widget,
       
       w = widget->requisition.width - OUTSIDE_SELECT_RECT * 2 - 1;
       h = widget->requisition.height - OUTSIDE_SELECT_RECT * 2 - 1;
-      
+
       gdk_draw_rectangle (widget->window,
                           widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
                           FALSE,
@@ -554,3 +605,198 @@ meta_select_image_expose_event (GtkWidget      *widget,
 
   return GTK_WIDGET_CLASS (parent_class)->expose_event (widget, event);
 }
+
+#define META_TYPE_SELECT_WORKSPACE   (meta_select_workspace_get_type ())
+#define META_SELECT_WORKSPACE(obj)   (GTK_CHECK_CAST ((obj), META_TYPE_SELECT_WORKSPACE, MetaSelectWorkspace))
+
+typedef struct _MetaSelectWorkspace       MetaSelectWorkspace;
+typedef struct _MetaSelectWorkspaceClass  MetaSelectWorkspaceClass;
+
+struct _MetaSelectWorkspace
+{
+  GtkDrawingArea parent_instance;
+  MetaWorkspace *workspace;
+  guint selected : 1;
+};
+
+struct _MetaSelectWorkspaceClass
+{
+  GtkDrawingAreaClass parent_class;
+};
+
+
+static GType meta_select_workspace_get_type (void) G_GNUC_CONST;
+
+#define SELECT_OUTLINE_WIDTH 2
+#define MINI_WORKSPACE_WIDTH 48
+
+static GtkWidget*
+selectable_workspace_new (MetaWorkspace *workspace)
+{
+  GtkWidget *widget;
+  double screen_aspect;
+  
+  widget = g_object_new (meta_select_workspace_get_type (), NULL);
+
+  screen_aspect = (double) workspace->screen->height / (double) workspace->screen->width;
+  
+  /* account for select rect */ 
+  gtk_widget_set_size_request (widget,
+                               MINI_WORKSPACE_WIDTH + SELECT_OUTLINE_WIDTH * 2,
+                               MINI_WORKSPACE_WIDTH * screen_aspect + SELECT_OUTLINE_WIDTH * 2);
+
+  META_SELECT_WORKSPACE (widget)->workspace = workspace;
+
+  return widget;
+}
+
+static void
+select_workspace (GtkWidget *widget)
+{
+  META_SELECT_WORKSPACE(widget)->selected = TRUE;
+  gtk_widget_queue_draw (widget);
+}
+
+static void
+unselect_workspace (GtkWidget *widget)
+{
+  META_SELECT_WORKSPACE (widget)->selected = FALSE;
+  gtk_widget_queue_draw (widget);
+}
+
+static void meta_select_workspace_class_init (MetaSelectWorkspaceClass *klass);
+
+static gboolean meta_select_workspace_expose_event (GtkWidget      *widget,
+                                                    GdkEventExpose *event);
+
+GType
+meta_select_workspace_get_type (void)
+{
+  static GtkType workspace_type = 0;
+
+  if (!workspace_type)
+    {
+      static const GTypeInfo workspace_info =
+      {
+        sizeof (MetaSelectWorkspaceClass),
+        NULL,           /* base_init */
+        NULL,           /* base_finalize */
+        (GClassInitFunc) meta_select_workspace_class_init,
+        NULL,           /* class_finalize */
+        NULL,           /* class_data */
+        sizeof (MetaSelectWorkspace),
+        16,             /* n_preallocs */
+        (GInstanceInitFunc) NULL,
+      };
+
+      workspace_type = g_type_register_static (GTK_TYPE_DRAWING_AREA, 
+                                               "MetaSelectWorkspace", 
+                                               &workspace_info, 
+                                               0);
+    }
+
+  return workspace_type;
+}
+
+static void
+meta_select_workspace_class_init (MetaSelectWorkspaceClass *klass)
+{
+  GtkWidgetClass *widget_class;
+  
+  widget_class = GTK_WIDGET_CLASS (klass);
+  
+  widget_class->expose_event = meta_select_workspace_expose_event;
+}
+
+static gboolean
+meta_select_workspace_expose_event (GtkWidget      *widget,
+                                    GdkEventExpose *event)
+{
+  MetaWorkspace *workspace;
+  WnckWindowDisplayInfo *windows;
+  int i, n_windows;
+  GList *tmp, *list;
+
+  workspace = META_SELECT_WORKSPACE (widget)->workspace;
+              
+  list = meta_stack_list_windows (workspace->screen->stack, workspace);
+  n_windows = g_list_length (list);
+  windows = g_new (WnckWindowDisplayInfo, n_windows);
+
+  tmp = list;
+  i = 0;
+  while (tmp != NULL)
+    {
+      MetaWindow *window;
+
+      window = tmp->data;
+
+      if (window->skip_pager || 
+          window->minimized || 
+          window->unmaps_pending)
+        {
+          --n_windows;
+        }
+      else
+        {
+          windows[i].icon = window->icon;
+          windows[i].mini_icon = window->mini_icon;
+          windows[i].is_active = window->has_focus;
+
+          if (window->frame)
+            {
+              windows[i].x = window->frame->rect.x;
+              windows[i].y = window->frame->rect.y;
+              windows[i].width = window->frame->rect.width;
+              windows[i].height = window->frame->rect.height;
+            }
+          else
+            {                
+              windows[i].x = window->rect.x;
+              windows[i].y = window->rect.y;
+              windows[i].width = window->rect.width;
+              windows[i].height = window->rect.height;
+            }
+
+          i++;
+        }
+      tmp = tmp->next;
+    }
+
+  g_list_free (list);
+
+  wnck_draw_workspace (widget,
+                       widget->window,
+                       SELECT_OUTLINE_WIDTH,
+                       SELECT_OUTLINE_WIDTH,
+                       widget->allocation.width - SELECT_OUTLINE_WIDTH * 2,
+                       widget->allocation.height - SELECT_OUTLINE_WIDTH * 2,
+                       workspace->screen->width,
+                       workspace->screen->height,
+                       NULL,
+                       (workspace->screen->active_workspace == workspace),
+                       windows,
+                       n_windows);
+
+  g_free (windows);
+  
+  if (META_SELECT_WORKSPACE (widget)->selected)
+    {
+      i = SELECT_OUTLINE_WIDTH - 1;
+      while (i >= 0)
+        {
+          gdk_draw_rectangle (widget->window,
+                              widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
+                              FALSE,
+                              i, 
+                              i,
+                              widget->allocation.width - i * 2 - 1,
+                              widget->allocation.height - i * 2 - 1);
+
+          --i;
+        }
+    }
+
+  return TRUE;
+}
+
