@@ -151,6 +151,37 @@ set_gtk_window_type (GtkWindow  *window,
                          type);
 }
 
+static void
+set_gdk_window_border_only (GdkWindow *window)
+{
+  gdk_window_set_decorations (window, GDK_DECOR_BORDER);
+}
+
+static void
+on_realize_set_border_only (GtkWindow *window,
+                            gpointer   data)
+{
+  g_return_if_fail (GTK_WIDGET_REALIZED (window));
+  
+  set_gdk_window_border_only (GTK_WIDGET (window)->window);
+}
+
+static void
+set_gtk_window_border_only (GtkWindow  *window)
+{
+  g_signal_handlers_disconnect_by_func (G_OBJECT (window),
+                                        on_realize_set_border_only,
+                                        NULL);
+                                
+  g_signal_connect_after (G_OBJECT (window),
+                          "realize",
+                          G_CALLBACK (on_realize_set_border_only),
+                          NULL);
+
+  if (GTK_WIDGET_REALIZED (window))
+    set_gdk_window_border_only (GTK_WIDGET (window)->window);
+}
+
 int
 main (int argc, char **argv)
 {
@@ -260,6 +291,125 @@ utility_cb (gpointer             callback_data,
   gtk_widget_show_all (window);
 }
 
+static void
+toolbar_cb (gpointer             callback_data,
+            guint                callback_action,
+            GtkWidget           *widget)
+{
+  GtkWidget *window;
+  GtkWidget *vbox;
+  GtkWidget *label;
+  
+  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
+  set_gtk_window_type (GTK_WINDOW (window), "_NET_WM_WINDOW_TYPE_TOOLBAR");
+  gtk_window_set_title (GTK_WINDOW (window), "Toolbar");
+  
+  gtk_window_set_transient_for (GTK_WINDOW (window), GTK_WINDOW (callback_data));
+  
+  vbox = gtk_vbox_new (FALSE, 0);
+
+  gtk_container_add (GTK_CONTAINER (window), vbox);
+
+  label = gtk_label_new ("FIXME this needs a resize grip, etc.");
+  gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
+  
+  gtk_widget_show_all (window);
+}
+
+static void
+menu_cb (gpointer             callback_data,
+         guint                callback_action,
+         GtkWidget           *widget)
+{
+  GtkWidget *window;
+  GtkWidget *vbox;
+  GtkWidget *label;
+  
+  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
+  set_gtk_window_type (GTK_WINDOW (window), "_NET_WM_WINDOW_TYPE_MENU");
+  gtk_window_set_title (GTK_WINDOW (window), "Menu");
+  
+  gtk_window_set_transient_for (GTK_WINDOW (window), GTK_WINDOW (callback_data));
+  
+  vbox = gtk_vbox_new (FALSE, 0);
+
+  gtk_container_add (GTK_CONTAINER (window), vbox);
+
+  label = gtk_label_new ("FIXME this isn't a menu.");
+  gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
+  
+  gtk_widget_show_all (window);
+}
+
+static void
+override_redirect_cb (gpointer             callback_data,
+                      guint                callback_action,
+                      GtkWidget           *widget)
+{
+  GtkWidget *window;
+  GtkWidget *vbox;
+  GtkWidget *label;
+  
+  window = gtk_window_new (GTK_WINDOW_POPUP);
+  gtk_window_set_title (GTK_WINDOW (window), "Override Redirect");
+  
+  vbox = gtk_vbox_new (FALSE, 0);
+
+  gtk_container_add (GTK_CONTAINER (window), vbox);
+
+  label = gtk_label_new ("This is an override\nredirect window\nand should not be managed");
+  gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
+  
+  gtk_widget_show_all (window);
+}
+
+static void
+border_only_cb (gpointer             callback_data,
+                guint                callback_action,
+                GtkWidget           *widget)
+{
+  GtkWidget *window;
+  GtkWidget *vbox;
+  GtkWidget *label;
+  
+  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
+  set_gtk_window_border_only (GTK_WINDOW (window));
+  gtk_window_set_title (GTK_WINDOW (window), "Border only");
+  
+  gtk_window_set_transient_for (GTK_WINDOW (window), GTK_WINDOW (callback_data));
+  
+  vbox = gtk_vbox_new (FALSE, 0);
+
+  gtk_container_add (GTK_CONTAINER (window), vbox);
+
+  label = gtk_label_new ("This window is supposed to have a border but no titlebar.");
+  gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
+  
+  gtk_widget_show_all (window);
+}
+
+static void
+changing_icon_cb (gpointer             callback_data,
+                  guint                callback_action,
+                  GtkWidget           *widget)
+{
+  GtkWidget *window;
+  GtkWidget *vbox;
+  GtkWidget *label;
+  
+  window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
+  gtk_window_set_title (GTK_WINDOW (window), "Changing Icon");
+  
+  vbox = gtk_vbox_new (FALSE, 0);
+
+  gtk_container_add (GTK_CONTAINER (window), vbox);
+
+  label = gtk_label_new ("This window has an icon that changes over time");
+  gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
+  
+  gtk_widget_show_all (window);
+}
+
 static gboolean
 focus_in_event_cb (GtkWidget *window,
                    GdkEvent  *event,
@@ -470,7 +620,11 @@ static GtkItemFactoryEntry menu_items[] =
   { "/Windows/_Left dock",    NULL,          dock_cb,                 DOCK_LEFT, NULL },
   { "/Windows/_Right dock",   NULL,          dock_cb,                 DOCK_RIGHT, NULL },
   { "/Windows/_All docks",    NULL,          dock_cb,                 DOCK_ALL, NULL },
-  { "/Windows/Des_ktop",      NULL,          desktop_cb,              0, NULL }
+  { "/Windows/Des_ktop",      NULL,          desktop_cb,              0, NULL },
+  { "/Windows/Me_nu",         NULL,          menu_cb,                 0, NULL },
+  { "/Windows/Tool_bar",      NULL,          toolbar_cb,              0, NULL },
+  { "/Windows/Override Redirect",      NULL,          override_redirect_cb,              0, NULL },
+  { "/Windows/Border Only",      NULL,          border_only_cb,              0, NULL }
 };
 
 static void
@@ -481,8 +635,8 @@ sleep_cb (GtkWidget *button,
 }
 
 static void
-toolbar_cb (GtkWidget *button,
-            gpointer   data)
+clicked_toolbar_cb (GtkWidget *button,
+                    gpointer   data)
 {
   GtkWidget *dialog;
   
@@ -629,7 +783,7 @@ do_appwindow (void)
                             GTK_STOCK_QUIT,
                             "This is a demo button with a 'quit' icon",
                             NULL,
-                            G_CALLBACK (toolbar_cb),
+                            G_CALLBACK (clicked_toolbar_cb),
                             window, /* user data for callback */
                             -1);  /* -1 means "append" */
 
