@@ -73,6 +73,11 @@ void meta_ui_unflicker_frame_bg (MetaUI *ui,
 void meta_ui_reset_frame_bg     (MetaUI *ui,
                                  Window  xwindow);
 
+void meta_ui_apply_frame_shape  (MetaUI *ui,
+                                 Window  xwindow,
+                                 int     new_window_width,
+                                 int     new_window_height);
+
 void meta_ui_queue_frame_draw (MetaUI *ui,
                                Window xwindow);
 
@@ -80,6 +85,9 @@ void meta_ui_set_frame_title (MetaUI *ui,
                               Window xwindow,
                               const char *title);
 
+void meta_ui_repaint_frame (MetaUI *ui,
+                            Window xwindow);
+
 MetaWindowMenu* meta_ui_window_menu_new   (MetaUI             *ui,
                                            Window              client_xwindow,
                                            MetaMenuOp          ops,
@@ -145,6 +153,9 @@ void     meta_ui_set_current_theme (const char *name,
                                     gboolean    force_reload);
 gboolean meta_ui_have_a_theme      (void);
 
+gboolean meta_ui_parse_accelerator (const char          *accel,
+                                    unsigned int        *keysym,
+                                    MetaVirtualModifier *mask);
 
 #include "tabpopup.h"
 
