@@ -29,6 +29,8 @@
 #include <X11/Xutil.h>
 #include <gdk-pixbuf/gdk-pixbuf.h>
 
+typedef void (*MetaWindowForeachFunc) (MetaWindow *window,
+                                       void       *data);
 
 typedef enum
 {
@@ -53,6 +55,7 @@ struct _MetaWindow
   MetaFrame *frame;
   int depth;
   Visual *xvisual;
+  Colormap colormap;
   char *desc; /* used in debug spew */
   char *title;
 
@@ -128,6 +131,7 @@ struct _MetaWindow
   
   /* MWM hints about features of window */
   guint mwm_decorated : 1;
+  guint mwm_border_only : 1;
   guint mwm_has_close_func : 1;
   guint mwm_has_minimize_func : 1;
   guint mwm_has_maximize_func : 1;
@@ -136,6 +140,8 @@ struct _MetaWindow
   
   /* Computed features of window */
   guint decorated : 1;
+  guint border_only : 1;
+  guint always_sticky : 1;
   guint has_close_func : 1;
   guint has_minimize_func : 1;
   guint has_maximize_func : 1;
@@ -146,12 +152,14 @@ struct _MetaWindow
   
   /* Weird "_NET_WM_STATE_MODAL" flag */
   guint wm_state_modal : 1;
-  /* If these are TRUE, it just means a client explicitly
-   * toggled them on; we compute actual _NET_WM_STATE from
-   * window type usually
-   */
+
+  /* TRUE if the client forced these on */
   guint wm_state_skip_taskbar : 1;
   guint wm_state_skip_pager : 1;
+
+  /* Computed whether to skip taskbar or not */
+  guint skip_taskbar : 1;
+  guint skip_pager : 1;
   
   /* this flag tracks receipt of focus_in focus_out and
    * determines whether we draw the focus
@@ -246,6 +254,14 @@ struct _MetaWindow
   int dialog_pipe;
 };
 
+#define META_WINDOW_ALLOWS_MOVE(w)     ((w)->has_move_func && !(w)->maximized && !(w)->fullscreen)
+#define META_WINDOW_ALLOWS_RESIZE_EXCEPT_HINTS(w)   ((w)->has_resize_func && !(w)->maximized && !(w)->fullscreen && !(w)->shaded)
+#define META_WINDOW_ALLOWS_RESIZE(w)   (META_WINDOW_ALLOWS_RESIZE_EXCEPT_HINTS (w) &&                \
+                                        (((w)->size_hints.min_width < (w)->size_hints.max_width) ||  \
+                                         ((w)->size_hints.min_height < (w)->size_hints.max_height)))
+#define META_WINDOW_ALLOWS_HORIZONTAL_RESIZE(w) (META_WINDOW_ALLOWS_RESIZE_EXCEPT_HINTS (w) && (w)->size_hints.min_width < (w)->size_hints.max_width)
+#define META_WINDOW_ALLOWS_VERTICAL_RESIZE(w)   (META_WINDOW_ALLOWS_RESIZE_EXCEPT_HINTS (w) && (w)->size_hints.min_height < (w)->size_hints.max_height)
+
 MetaWindow* meta_window_new                (MetaDisplay *display,
                                             Window       xwindow,
                                             gboolean     must_be_viewable);
@@ -368,6 +384,7 @@ gboolean meta_window_visible_on_workspace (MetaWindow    *window,
 
 /* Get minimum work area for all workspaces we're on */
 void meta_window_get_work_area (MetaWindow    *window,
+                                gboolean       for_current_xinerama,
                                 MetaRectangle *area);
 
 gboolean meta_window_same_application (MetaWindow *window,
@@ -383,4 +400,10 @@ void meta_window_refresh_resize_popup (MetaWindow *window);
 
 void meta_window_free_delete_dialog (MetaWindow *window);
 
+void meta_window_foreach_transient (MetaWindow            *window,
+                                    MetaWindowForeachFunc  func,
+                                    void                  *data);
+gboolean meta_window_is_ancestor_of_transient (MetaWindow *window,
+                                               MetaWindow *transient);
+
 #endif
